/* $Novell: ssl_init.c,v 1.56 2003/05/13 13:38:03 $ 
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ******************************************************************************
 * 
 *	Copyright (c) 2004 Novell, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without 
 *	modification, are permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice, this 
 *	list of conditions and the following disclaimer. 
 *
 *	Redistributions in binary form must reproduce the above copyright notice, 
 *	this list of conditions and the following disclaimer in the documentation 
 *	and or other materials provided with the distribution.
 *
 *	Neither the name of Novell nor the names of its contributors may be used to 
 *	endorse or promote products derived from this software without specific prior
 *	written permission. 
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 *	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 *	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *	POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************
 */

/****************************************************************************
*
* Description:   This file contains initialization functions for configuring
*  the LDAP SDK to use SSL
*
****************************************************************************/
#include <stdio.h>

#if defined(N_PLAT_WIN32) || defined(WIN64)
   #include <windows.h>

#elif defined(MODESTO)
   #include "nksapi.h"
#endif

#include "ldap_ssl.h"

#include "ldap.h"

#include "openssl/ssl.h"
#include "openssl/rand.h"
#include "openssl/err.h"
#include "openssl/x509.h"
#include "openssl/x509_vfy.h"
#include "ssl_cert.h"

#include "globals.h"
#include "debug.h"

/****************************************************************************
*
* Description:
*
*  ldapssl_client_init initializes the SSL library.  The
*  function creates a "master" crypto context that is "cloned" and
*  bound to a socket during the prepare function.  The creation of the
*  global cypto context is synchronized to ensure only one thread is
*  creating it at a time.
*
*  NOTE: This api allows clients to call it multiple times ( reinitialize
*        the ssl stuff ).  Doing this will create a new global crypto
*        context and initialize it with the new information.  This does
*        not invalidate any existing ssl connections.  Existing connections
*        will continue to function and use the crypto context that was
*        created during the iof_prepare function.  However, new connections
*        will use the new crytpo context.
****************************************************************************/

LDAP_F(int) ldapssl_client_init
(
   const char *certFile,
   void       *reserved
)
{
   int         rc = LDAPSSL_SUCCESS;
   SSL_CTX     *newCtx;
   X509_LOOKUP *lookup;
   int         sslReg = 0;

#ifdef UNIX
	FILE	*fh;
	static const size_t RAND_SIZE = 160/8;
#endif

   DBG_FUNC("ldapssl_client_init");
   DBG_PARM("certFile: %s\n", certFile ? certFile : "NULL ptr" );
   DBG_LINE();

   EnterSync();
   sslReg = GET_SSL_REGISTERED;
   ExitSync();
      
      /* initialize the OpenSSL library
      */
   if (!sslReg)
   {
      rc = SSL_library_init();
      DBG_OPENSSL_ERROR(rc);

      if (1 == rc) /* SSL_library_init() returns 1 if successful */
      {
         rc = LDAPSSL_SUCCESS;
         EnterSync();
         SET_SSL_REGISTERED;
         ExitSync();
      }
      else
      {
         DBG_RETURN(LDAPSSL_ERROR);
      }

         /* if building debug load the error strings for better
          * error messages
         */
      #if defined (DEBUG)
      SSL_load_error_strings();
      #endif

         /* Add PRNG seed material - everytime an app calls init we will
          * add more data.
         */
      #ifndef UNIX 
      RAND_poll();    
      #endif /* if !(UNIX) */

      #ifdef UNIX
	  /* Use a random entropy pool device if available otherwise use 
		 BBS alogrithem to get the random data.  Linux, FreeBSD and 
		 OpenBSD have entropy pool device.  Use /dev/random if you can
		 as /dev/random may block if it runs out of random entries */

		 if ((fh = fopen("/dev/urandom", "r"))!=NULL)
		 {
			setvbuf(fh, NULL, _IONBF, 0);
			randbuf=(unsigned char *)malloc(RAND_SIZE);
			fread((unsigned char *)randbuf,1,RAND_SIZE,fh);
			fclose(fh);
			RAND_seed(randbuf,sizeof(randbuf));
         }
		 else
         if ((fh = fopen("/dev/random", "r"))!=NULL)
         {
            setvbuf(fh, NULL, _IONBF, 0);
			randbuf=(unsigned char *)malloc(RAND_SIZE);
			fread((unsigned char *)randbuf,1,RAND_SIZE,fh);
			fclose(fh);
			RAND_seed(randbuf,sizeof(randbuf));
         }
/*
		 else
		 {
             randbuf = getRandomData(RAND_SIZE);
             RAND_seed(randbuf,sizeof(randbuf)); 
         }
*/
		 free(randbuf);
      #endif /* ifdef UNIX */

   } /* end if !SSLreg */

      /* create a new master crypto context and setup intial defaults
      */
   newCtx = SSL_CTX_new( SSLv23_method() );
   DBG_OPENSSL_ERROR( (NULL != newCtx ? 1 : 0 ) );

   if (NULL != newCtx)
   {
         /* turn off TLS so we work with the IBM ldap directory
         */
      SSL_CTX_set_options(newCtx, SSL_OP_NO_TLSv1);

         /* default verify mode to server certificate verification
          * NOTE: interative cert verification isn't working on UNIX 
         */
      SSL_CTX_set_verify(newCtx, SSL_VERIFY_PEER, NULL);


         /* if a trusted root certificate file is passed in load it into
          * the master crypt ctx - NOTE: the files passed in to this
          * api must be "DER" encoded.  We require this for backwards 
          * compatibility.  Originally we only accepted "der" files.  Now
          * applications can also use "base64" encoded files but they
          * must use the new API (ldapssl_add_trusted_cert).
         */
      if (NULL != certFile)
      {
         lookup = X509_STORE_add_lookup(newCtx->cert_store, 
                                        X509_LOOKUP_file());

         DBG_OPENSSL_ERROR( (NULL != lookup ? 1 : 0 ) );

         if (NULL != lookup)
         {
            rc = X509_load_cert_file( lookup, certFile, X509_FILETYPE_ASN1);
            DBG_OPENSSL_ERROR(rc);

            if (1 == rc) /* openSSL returns 1 for success */
               rc = LDAPSSL_SUCCESS;
            else
               rc = LDAPSSL_ERROR;
         }
         else
            rc = LDAPSSL_ERROR;
      }

         /* Support to specify ciphers - cipher string is passed in
            through the reserved parm - not normally available
         */
      #if defined(CONFIG_CIPHERS)
      DBG_OUTPUT("%s\n", "Configurable Ciphers build");
      if (NULL != reserved)
      {
         DBG_OUTPUT("  Ciphers: %s\n", (char*)reserved);
         rc = SSL_CTX_set_cipher_list(newCtx, (char*)reserved);
         if (1 == rc)
            rc = LDAPSSL_SUCCESS;
         else
            rc = LDAPSSL_ERROR;
      }
      #endif
   }
   else
   {
      rc = LDAPSSL_ERROR;
   }  /* if (NULL != newCtx) */


      /* successfully created a new master crypt context 
      */
   if (LDAPSSL_SUCCESS == rc )
   {
      EnterSync();

         /* free existing master crypt context - the first time
          * the master crypt ctx will be NULL but it won't matter
         */
      SSL_CTX_free( GetMasterCryptCtx() );
      
      SetMasterCryptCtx(newCtx);

      ExitSync();
   }
   else /* if there was an error free the new context */
   {
      if (NULL != newCtx)
        SSL_CTX_free( newCtx );
   }
   
   ldap_set_option(NULL,LDAP_OPT_X_TLS_CTX,newCtx);
   DBG_RETURN(rc);
}

/****************************************************************************
*
* Description:
*
*  ldapssl_client_deinit deinitializes the SSL library
*
****************************************************************************/
LDAP_F(int) ldapssl_client_deinit
(
   void
)
{
   SSL_CTX *mCryptCtx;

#ifdef LDAPSSL_DEBUG
   DBG_FUNC("ldapssl_client_deinit");
   DBG_LINE();
#endif /* LDAPSSL_DEBUG */

      /* free the master cryto context
      */
   EnterSync();

   if (!GET_SSL_REGISTERED)
   {
      ExitSync();
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: SSL library not initialized\n");
      DBG_RETURN(LDAPSSL_ERROR);
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);
   }

   mCryptCtx = GetMasterCryptCtx();

   if (NULL != mCryptCtx )
   {
      SSL_CTX_free( mCryptCtx );
      SetMasterCryptCtx(NULL);
   }                                

   SetCertCallback(NULL);
   SET_SSL_UNREGISTERED;
   ExitSync();


      /* If code doesn't exit early due to ssl unregistered, always return
       * success.  Could return SAS or WinSock SSL errors but they aren't
       * really valuable to the application.
      */
#ifdef LDAPSSL_DEBUG
   DBG_RETURN(LDAPSSL_SUCCESS);
#endif /*LDAPSSL_DEBUG*/
   return(LDAPSSL_SUCCESS);
}





