/*
 * $Novell: lcidentity.c,v 1.24 2002/09/26 19:54:14 $
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************
 *
 * Copyright (C) 1999, 2000 Novell, Inc. All Rights Reserved.
 * 
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 
 ******************************************************************************
 */

#if defined(WIN32)
#include<windows.h>
#define uint16 unsigned short
#define uint8 unsigned char

#elif defined(UNIX)
#define uint16 unsigned short
#define uint8  unsigned char

#elif defined(MODESTO)
#include <stdlib.h>
#define uint16 unsigned short
#define uint8  unsigned char

#endif

//#include "portable.h"

/* Core LDAP header                                         */
#include "ldap.h"

/* The SDK header file                                      */
#include "ldapx.h"

/* Win32 specific stuff goes here                           */
#if defined (WIN32)
#define ldap_wcslen wcslen
#else
#define ldap_wcslen x_unilen
#endif

#ifdef HAVE_MALLOC_H 
#include <malloc.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#if defined(UNIX) || defined(MODESTO)
static size_t x_unilen(unsigned short *s)
{
   unsigned i = 0;
   while (*s) {i++; s++;}
   return i;
}
#endif  

/*
 *  Conversion Routine for Big Endian (host independent conversion
 */
static uint16 getint16(uint16 num)
{
    /* The following blows the Watcom compiler
    * So we must do a less efficient implementation
    * return(   ( ((uint8 *)(&(num)))[0] << 8 ) | ( ((uint8 *)(&(num)))[1] ));
    */
    uint8 *cn = (uint8 *)&num;
    return( (cn[0] << 8 ) | cn[1] );
}
#define GETINT16(num) getint16(num)
#define PUTINT16(num) getint16(num)


/* ------------------------------------------------------------------
ldap_get_bind_dn
==============================

RequestBer
    Verb        INTEGER

ResponseBer
    identity    OCTET STRING

Description
    This API returns the LDAP name of the object we are logged in as
-------------------------------------------------------------------- */
LDAP_F(int) ldap_get_bind_dn(LDAP* ld, char** identity)
{
    BerElement*     returnedber = NULL;
    int             err = LDAP_SUCCESS;
    ber_len_t       len = 0;

    struct berval*  identitybv = NULL;

    char*           returnedOID = NULL;
    struct berval*  returnedbv = NULL;


    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_GET_BIND_DN_EXTENDED_REQUEST, NULL, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;

    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }


    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_GET_BIND_DN_EXTENDED_REPLY)) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }

    /* Do we have a good returned berval */
    if (!returnedbv) {
        /* No returned berval means rather drastic error. Return operations error */
        err = LDAP_OPERATIONS_ERROR;
        goto cleanup;   
    }


    /* Build a berElement from the returned berval */
    returnedber = ber_init(returnedbv);

    if (returnedber == NULL) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* The returned data contains the identiy - scan it */
    err = ber_scanf(returnedber, "O", &identitybv);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    } else {

        /* Note that the caller frees the identity pointer when done */
        *identity = (char *)malloc( identitybv->bv_len + 1);
        if (!(*identity)) {
            err = LDAP_NO_MEMORY;
            goto cleanup;
        }

        /* Use the smaller of the two sizes. Though highly unlikely trying
         * to guard against a size over-run */
        len = (L_MAX_DN_BYTES < identitybv->bv_len) ? L_MAX_DN_BYTES - 1 : identitybv->bv_len;
        memcpy( *identity, identitybv->bv_val, len);
        
        /* Null terminate the string */
        *(*identity + len) = '\0';
        ber_bvfree(identitybv);

        /* Extended request completed successfully */
        err = LDAP_SUCCESS;
    }

cleanup:
    if (returnedber)
        ber_free(returnedber, 1);

    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* return the appropriate error/success code */
    return (err); 
}
LDAP_F(int) ldap_get_context_identity_name(LDAP* ld, char** identity)
{
    return ldap_get_bind_dn(ld, identity);
}


/* ------------------------------------------------------------------
ldap_get_effective_privileges
==============================

RequestBer
    verb        INTEGER
    dn          OCTET STRING
    trusteeDN   OCTET STRING
    attrName    OCTET STRING

ResponseBer
    privileges  INTEGER

Description
    This API returns the rights that trusteeDN has on the attrName 
    attribute.  For a list of returned privileges see the ldapx.h 
    header file.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_get_effective_privileges(LDAP* ld, char* dn, char* trusteeDN, char* attrName, int* privileges)
{

    BerElement*    requestber = NULL;
    BerElement*    returnedber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;


    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the dn, the trusteeDN and attrname parameters */
    err = ber_printf(requestber, "sss", dn, trusteeDN, attrName);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto done;
    }

    /* Converts the BER we just built to a berval that we will send out
    in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto done;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_GET_EFFECTIVE_PRIVILEGES_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto done;

    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto done;
    }

    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_GET_EFFECTIVE_PRIVILEGES_EXTENDED_REPLY)) {
        err = LDAP_NOT_SUPPORTED;
        goto done;
    }

    /* Do we have a good returned berval */
    if (!returnedbv) {
        /* No returned berval means rather drastic error. Return operations error */
        err = LDAP_OPERATIONS_ERROR;
        goto done;
    }

    /* Build a berElement from the returned berval */
    returnedber = ber_init(returnedbv);
    if (returnedber == NULL) {
        err = LDAP_NO_MEMORY;
        goto done;
    }

    /* The returned data contains the privilege flag - scan it */
    err = ber_scanf(returnedber, "i", privileges);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto done;
    }
    else
        /* Extended request completed successfully */
        err = LDAP_SUCCESS;

done:
    if (returnedber)
        ber_free(returnedber, 1);

    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1);

    /* return the appropriate error/success code */
    return (err);
}


/* ------------------------------------------------------------------
ldap_nds_to_ldap
==============================

RequestBer
    verb        INTEGER
    dn          OCTET STRING

ResponseBer
     ldapName   OCTET STRING

Description
  This API is provided for legacy NDS applications. It converts a 
  dot-sperated-Unicode NDS name to a typed LDAP name. Caller 
  has the responsibility of freeing ldapName by calling ldap_memfree
-------------------------------------------------------------------- */
LDAP_F(int) ldap_nds_to_ldap(LDAP* ld, unsigned short* dn, char** ldapName)
{
    BerElement      *requestber = NULL;
    BerElement      *returnedber = NULL;
    int             err = LDAP_SUCCESS;
    struct berval   *requestbv = NULL;

    char            *returnedOID = NULL;
    struct berval   *returnedbv = NULL;
    struct berval   *ldapNamebv = NULL;
    size_t           byteSize = 0;
    size_t           dnSize = 0;
    uint16          *bigEndianDN = NULL;
    uint16          *tempPtr = NULL;
    size_t           i = 0;
   
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* dn is unicode by design */
    byteSize = ldap_wcslen(dn);
    dnSize = ( byteSize + 1 ) * 2;

    bigEndianDN = (uint16 *) malloc (dnSize);
    if (!bigEndianDN)
        return LDAP_NO_MEMORY;

    // Loop through the characters and stuff a big endian double
    // byte into the bigEndianDN
    tempPtr = bigEndianDN;
    for (i = 0; ( (i < byteSize) && (dn) && (tempPtr) ); i++) {
        *tempPtr = PUTINT16(*dn);
        dn++;
        tempPtr++;
    }
    // Null terminate the big endian string
    *tempPtr = 0;
   
    /* stuff in the Verb, the NDS distinguished Name and name length */
    err = ber_printf(requestber, "o", bigEndianDN, dnSize);
    
    // Irrespective of what happened to printf first free memory
    if (bigEndianDN)
        free (bigEndianDN);

    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
      in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_NDS_TO_LDAP_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_NDS_TO_LDAP_EXTENDED_REPLY)) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;   
    }

    /* Do we have a good returned berval */
    if (!returnedbv) {
        /* No returned berval means rather drastic error. Return operations error */
        err = LDAP_OPERATIONS_ERROR;
        goto cleanup;   
    }

      
    /* Build a berElement from the returned berval */
    returnedber = ber_init(returnedbv);
    if (returnedber == NULL) {
       err = LDAP_NO_MEMORY;
       goto cleanup;
    }

    /* The returned data contains the ldap name - scan it */
    err = ber_scanf(returnedber, "O", &ldapNamebv);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }
    else {
        /* Caller frees ldapName, this API allocates it for you*/
        *ldapName = (char *)malloc( ldapNamebv->bv_len + 1);
        if (!(*ldapName)) {
            err = LDAP_NO_MEMORY;
            goto cleanup;
        }

        memcpy( *ldapName, ldapNamebv->bv_val, ldapNamebv->bv_len);

        *(*ldapName + ldapNamebv->bv_len) = '\0';
        ber_bvfree( ldapNamebv);
         
        /* Extended request completed successfully */
        err = LDAP_SUCCESS;
    }
   
cleanup:
    if (returnedber)
        ber_free(returnedber, 1);

    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1);  

   
    /* return the appropriate error/success code */
    return (err);   
}

/* ------------------------------------------------------------------
ldap_refresh_server
==============================

RequestBer

ResponseBer


Description
    This API refreshes the ldap server assuming you have rights.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_refresh_server(LDAP* ld)
{
   int            err = LDAP_SUCCESS;

   char*       returnedOID = NULL;
   struct berval* returnedbv = NULL;
   
   /* Call the ldap_extended_operation (synchronously) */
   if ((err = ldap_extended_operation_s(ld, NLDAP_REFRESH_SERVER_REQUEST, NULL, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
      goto done;
   
   /* Make sure there is a returned OID */
   if (!returnedOID) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }
   
   /* Is this what we were expecting to get back */
   if (strcmp(returnedOID, NLDAP_REFRESH_SERVER_REPLY)) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }

done:
   /* Free the returned OID string if one was returned */
   if (returnedOID)
       ldap_memfree(returnedOID);

   /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
      ber_bvfree(returnedbv);

   /* return the appropriate error/success code */
   return (err);   
}


/* ------------------------------------------------------------------
ldapx_memfree
==============================

RequestBer

ResponseBer


Description
    Applications must use this API is to free any mmeory that 
    the library allocates.
-------------------------------------------------------------------- */
LDAP_F(void) ldapx_memfree(void* mem)
{

    if (mem)
        free (mem);
}
