/*
 * $Novell: schema_util.c,v 1.28 2002/05/23 22:50:33 $
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ******************************************************************************
 * Copyright (C) 2001 Novell, Inc. All Rights Reserved.
 * 
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 
 *
 * This file is an implementation of the IETF draft,
 *	draft-morris-ldap-c-api-schema-01.txt
 ******************************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <assert.h>

#include "time.h"
#include "ctype.h"

#include "ldap.h" 
#include "ldap_schema.h"


struct ldap_schema_named_element {
    struct ldap_schema_element *p_element;
    char* name; 
};


/* helper functions for schema_fetch & schema_free: */
int get_schema_dn ( LDAP *ld, char** p_schemaDN );
int alloc_list    ( LDAPSchema *schema, int type, int count );
int parse_element ( LDAPSchema *schema, char* rawValue, int type );
int add_element   ( LDAPSchema *schema, LDAPSchemaElement *element );
int free_element  ( LDAPSchemaElement *element );
void free_changes ( LDAPSchema *schema );

/* used for qsort in fetch */
int compare_oid ( const void* element1, const void *element2 );
int compare_name( const void* element1, const void *element2 );

/* used for bsearch in ldap_schema_get_by_name */
int compare_key_to_oid ( const void* key, const void *schemaElement );
int compare_key_to_name( const void* key, const void *namedElement  );

/* string compare, Case Insensitive */
int strcasecmp( const char* a, const char* b);

/* helper functions for get functions for string and stringArray */
char **  str_array_dup( char *stringArray[] );

/* helper functions of ldap_get_field_values */
int single_string_dup( char *(*newValue[]), char *toDup   );
int double_string_dup( char *(*newValue[]), char *toDup[] );

/* helper functions of ldap_schema_add, modify, delete */
int ldap_schema_element_to_string( LDAPSchemaElement* element, char *str[], 
                                   char *typeString[] );
int store_update ( LDAPSchema *schema, LDAPSchemaElement * el, int modop );
int dup_element  ( LDAPSchemaElement *element, 
                    LDAPSchemaElement *newElement[] );
int modify_element( LDAPSchemaElement *element, char *fieldName, 
                    char *values[], int modType);
int remove_element( LDAPSchema* schema, char* oid, int elementType);
int new_values( char *(*p_old[]), char *values[], int modType);
int new_value ( char *p_str[], char *values[], int modType);
int add_change( LDAPSchema *schema, LDAPMod *change);

/* keyword in ldap_schema to denote a valid struct. */
#define KEYWORD 0xdeadbeef  


/**********************************************************************
    ldap_schema_fetch:
    Reads the Root DSE to find the location of schema definitions. It 
    retrieves all schema element definitions and stores them in a list,
    sorted by oid.  In addition it creates another list of names, 
    pointing to the schema elements, sorted by name.
 */    
LDAP_F( int )
ldap_schema_fetch( LDAP *ld, LDAPSchema **schema, 
    const char* subschemaSubentryDN )
{
    char        *attribute = NULL, *schemaDN = NULL, **values = NULL;
    BerElement  *ber = NULL;
    LDAPMessage *searchResult = NULL, *entry = NULL;
    int rc = 0,i = 0, count = 0;
    char        *attributes[] = { ATTRIBUTE_TYPES, OBJECT_CLASSES, 
                    LDAP_SYNTAXES, MATCHING_RULES, MATCHING_RULE_USE, 
                    DIT_CONTENT_RULES, DIT_STRUCTURE_RULES, NAME_FORMS, NULL };

        /* Calloc will zero out memory */
    *schema = LDAP_CALLOC(1, sizeof(LDAPSchema));
    if (*schema == NULL)
        return LDAP_NO_MEMORY;

	/* initialize variables */
	(*schema)->changes = NULL;
	(*schema)->changesCount = 0;
	(*schema)->changesAllocated = 0;

    if (subschemaSubentryDN == NULL)
    {
        rc = get_schema_dn( ld , &schemaDN );

        /* could not read the root DSE */
        if (rc != LDAP_SUCCESS) {
            LDAP_FREE(*schema);
            return rc;
        }
    }
    else schemaDN = (char*) subschemaSubentryDN;

    /* Search the subschemaSubentry*/
    rc = ldap_search_s(ld, schemaDN, LDAP_SCOPE_BASE, 
            "(objectclass=*)", attributes , 0, &searchResult); 

    if (subschemaSubentryDN == NULL)
        LDAP_FREE( schemaDN );

    if (rc != LDAP_SUCCESS) {
        ldap_msgfree( searchResult );
        LDAP_FREE(*schema);
        return rc;
    }

    /* Go through the search results */
    /* There should only be one entry */
    if ( (entry = ldap_first_entry( ld, searchResult )) != NULL ) 
    {
        for (   attribute =  ldap_first_attribute( ld, entry, &ber );
                attribute != NULL; 
                attribute =  ldap_next_attribute( ld, entry, ber ) ) {
            values = (char **)ldap_get_values( ld, entry, attribute );
            if (values == NULL) 
                continue;
            count  = ldap_count_values( values );            

            if ((strcmp( ATTRIBUTE_TYPES, attribute ) == 0) && 
                ((rc = alloc_list( *schema, LDAP_SCHEMA_ATTRIBUTE_TYPE, 
                                   count)) == LDAP_SUCCESS)) 
            {
                for (i = 0; values[i] != NULL; i++ ) 
                {
                    rc = parse_element( *schema, values[i], 
                                        LDAP_SCHEMA_ATTRIBUTE_TYPE);
                    if ( rc != LDAP_SUCCESS ) {
                        /* temporary workaround for invalid attributes 
                        break;  */
                        rc = LDAP_SUCCESS;
                    }
                }
            } else if ((strcmp( OBJECT_CLASSES, attribute ) == 0) && 
                       ((rc = alloc_list( *schema, LDAP_SCHEMA_OBJECT_CLASS, 
                                          count)) == LDAP_SUCCESS)) 
            {
                for (i = 0; values[i] != NULL; i++ ) 
                {
                    rc = parse_element( *schema, values[i], 
                                        LDAP_SCHEMA_OBJECT_CLASS );
                    if (rc != LDAP_SUCCESS) {
                        /* temporary workaround for invalid objects 
                        break;  */
                        rc = LDAP_SUCCESS;
                    }
                }
            } else if ((strcmp( LDAP_SYNTAXES, attribute ) == 0) &&    
                       ((rc = alloc_list( *schema, LDAP_SCHEMA_SYNTAX, 
                                          count)) == LDAP_SUCCESS)) 
            {
                for (i = 0; values[i] != NULL; i++ )
                {
                    rc = parse_element( *schema, values[i], 
                                        LDAP_SCHEMA_SYNTAX);
                    if (rc != LDAP_SUCCESS) {
                        break;
                    }
                }
            } else if ((strcmp( MATCHING_RULES, attribute ) == 0) && 
                       ((rc = alloc_list( *schema, LDAP_SCHEMA_MATCHING_RULE, 
                                          count)) == LDAP_SUCCESS)) 
            {
                for (i = 0; values[i] != NULL; i++ ) 
                {
                    rc = parse_element( *schema, values[i], 
                                        LDAP_SCHEMA_MATCHING_RULE);
                    if (rc != LDAP_SUCCESS) {
                        break;
                    }
                }
            } else if ((strcmp( MATCHING_RULE_USE, attribute ) == 0) && 
                       ((rc = alloc_list( *schema, LDAP_SCHEMA_MATCHING_RULE_USE, 
                                          count)) == LDAP_SUCCESS)) 
            {
                for (i = 0; values[i] != NULL; i++ ) 
                {
                    rc = parse_element( *schema, values[i], 
                                        LDAP_SCHEMA_MATCHING_RULE_USE);
                    if (rc != LDAP_SUCCESS)
                        break;
                }
            } else if ((strcmp( DIT_CONTENT_RULES, attribute ) == 0) && 
                       ((rc = alloc_list( *schema, LDAP_SCHEMA_DIT_CONTENT_RULE, 
                                          count)) == LDAP_SUCCESS)) 
            {
                for (i = 0; values[i] != NULL; i++ ) 
                {
                    rc = parse_element( *schema, values[i], 
                                        LDAP_SCHEMA_DIT_CONTENT_RULE);
                    if (rc != LDAP_SUCCESS)
                        break;
                }
            } else if ((strcmp( DIT_STRUCTURE_RULES, attribute ) == 0) && 
                       ((rc = alloc_list( *schema, LDAP_SCHEMA_DIT_STRUCTURE_RULE, 
                                          count)) == LDAP_SUCCESS)) 
            {
                for (i = 0; values[i] != NULL; i++ ) 
                {
                    rc = parse_element( *schema, values[i], 
                                        LDAP_SCHEMA_DIT_STRUCTURE_RULE);
                    if (rc != LDAP_SUCCESS)
                        break;
                }
            } else if ((strcmp( NAME_FORMS, attribute ) == 0) && 
                       ((rc = alloc_list( *schema, LDAP_SCHEMA_NAME_FORM, 
                                          count)) == LDAP_SUCCESS)) 
            {
                for (i = 0; values[i] != NULL; i++ ) 
                {
                    rc = parse_element( *schema, values[i], 
                                        LDAP_SCHEMA_NAME_FORM);
                    if (rc != LDAP_SUCCESS)
                        break;
                }
            }

            ldap_value_free( values );
            ldap_memfree( attribute );
            
            if (rc != LDAP_SUCCESS) {
                ldap_schema_free(*schema);
                return rc;
            }             
        }
        ber_free( ber, 0 );
    }
    ldap_msgfree( searchResult );

    /* sort each list for binary lookup */
    for ( i  = LDAP_SCHEMA_ATTRIBUTE_TYPE;
        i <= LDAP_SCHEMA_DIT_STRUCTURE_RULE;
        i ++) {
        qsort(  (void*)(*schema)->oids[i],          /* list to sort         */
                (*schema)->oidCount[i],             /* number of elements   */
                sizeof( LDAPSchemaElement*),        /* size of each item    */
                compare_oid  );                     /* Comparison function  */
        qsort(  (void*)(*schema)->names[i],
                (*schema)->nameCount[i],
                sizeof( struct ldap_schema_named_element ),
                compare_name );
    }
    (*schema)->testInit = KEYWORD;
    return( rc );
}



/**********************************************************************
    ldap_schema_free:
    frees the memory allocated by ldap_schema_fetch 
 */

LDAP_F( int )
ldap_schema_free( LDAPSchema *schema){
    int i,j; 

    if (schema == NULL)
        return LDAP_SUCCESS;
    if (schema->testInit != KEYWORD)
        return LDAP_PARAM_ERROR;
    schema->testInit = 0;

    for (i = LDAP_SCHEMA_ATTRIBUTE_TYPE; 
         i<= LDAP_SCHEMA_DIT_STRUCTURE_RULE; i++) {
        if (schema->names[i] != NULL) {
            for ( j=0; j < schema->nameCount[i]; j++)
                if (schema->names[i][j].name != NULL)
                    LDAP_FREE(schema->names[i][j].name);
            LDAP_FREE( schema->names[i] );
        }

        if (schema->oids[i] != NULL) {
            for ( j=0; j < schema->oidCount[i]; j++) {
                free_element( schema->oids[i][j] );                
            }
            LDAP_FREE( schema->oids[i] );
        }
    }
    free_changes(schema);
    LDAP_FREE( schema );            
    return LDAP_SUCCESS;
}


/*********************************************************************
    free_element:
    used to free memory allocated to an LDAPSchemaElement:
    1)element 2) LDAPSchemaElement struct
 */
int free_element( LDAPSchemaElement *element )
{
    switch (element->type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        ldap_attributetype_free(element->element.attribute);
        break;
    case LDAP_SCHEMA_OBJECT_CLASS:
        ldap_objectclass_free(element->element.object);
        break;
    case LDAP_SCHEMA_SYNTAX:
        ldap_syntax_free(element->element.syntax);
        break;
    case LDAP_SCHEMA_MATCHING_RULE:
        ldap_matchingrule_free(element->element.matching);
        break;
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        ldap_matchingruleuse_free(element->element.matchinguse);
        break;
    case LDAP_SCHEMA_NAME_FORM:
        ldap_nameform_free(element->element.nameform);
        break;
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        ldap_ditcontentrule_free(element->element.ditcontent);
        break;
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:
        ldap_ditstructurerule_free(element->element.ditstructure);
        break;
    default:
        assert(0);
        return LDAP_PARAM_ERROR;
    }
    
    LDAP_FREE( element );
    return LDAP_SUCCESS;
}


/*********************************************************************
    get_schema_dn:
    Reads the root DSE to locate the subschemaSubentry.
    Memory is allocated to the string schemaDN and must be freed by 
    LDAP_FREE();
    Note that the version MUST be set to v3.  
 */
int get_schema_dn( LDAP *ld, char** schemaDN ){
    char        *attribute = NULL, **values = NULL;    
    BerElement  *ber;
    LDAPMessage *searchResult = NULL, *entry;
    char *attributes[] = {"subschemaSubentry" , NULL};
    int rc;  

    *schemaDN = NULL;
    /* Search the Root DSE for subschemaSubentry */
    rc = ldap_search_ext_s(ld,
                           "",                 //Base DN
                           LDAP_SCOPE_BASE,    //Scope
                           "(objectclass=*)",  //filter
                           attributes,         //
                           0,                  //return attribute values
                           NULL,               //server Controls
                           NULL,               //client Controls
                           NULL,               //time out
                           LDAP_NO_LIMIT,      //NO limit on entries
                           &searchResult);     //results

    if ( rc != LDAP_SUCCESS ) {
        ldap_msgfree( searchResult );
        return rc;
    }

    rc = LDAP_SUCCESS;
    if (((entry      = ldap_first_entry( ld, searchResult    )) == NULL )||
        ((attribute  = ldap_first_attribute( ld, entry, &ber )) == NULL )|| 
        ((values     = (char **)ldap_get_values( ld, entry, attribute )) == NULL )) {   
            //there should only be one entry, one attribute, and one value
        rc = LDAP_NO_RESULTS_RETURNED;         
    } else {
        *schemaDN = LDAP_STRDUP(values[0]);
        if (*schemaDN == NULL)
            rc = LDAP_NO_MEMORY;
    }

    if (values != NULL)
       ldap_value_free( values );

    if (attribute != NULL)
       ldap_memfree( attribute );

    ldap_msgfree( searchResult );

    ber_free(ber, 0);

    return rc;
}


/*********************************************************************
    alloc_list:
    Used exclusively in ldap_schema_fetch, alloc_list allocates memory 
    for the list of oids and list of names.  Returns an LDAP result code 
*/
int alloc_list( LDAPSchema *schema, int type, int count)
{    
    schema->oids[type] = LDAP_CALLOC( count, sizeof(LDAPSchemaElement *));
    if (schema->oids[type] == NULL)
        return LDAP_NO_MEMORY;
    schema->oidsAllocated[type] = count;

    /* allocate name list. Note that count is just an estimate */
    schema->names[type] = 
        LDAP_CALLOC( count, sizeof( struct ldap_schema_named_element ));
    if (schema->names[type] == NULL)
        return LDAP_NO_MEMORY;
    schema->namesAllocated[type] = count;
    return LDAP_SUCCESS;
}



/*********************************************************************
    parse_element:
    Used exclusively in ldap_schema_fetch.  Uses routines in schema.c 
    to parse the rawValue and allocate space for the new element.
 */
int parse_element( LDAPSchema *schema, char* rawValue, int type ){
    LDAPSchemaElement *p_element;
    const char *errorMessage;
    int rc = LDAP_SUCCESS;    

    p_element = LDAP_MALLOC( sizeof(LDAPSchemaElement));
    if (p_element == NULL)
        return LDAP_NO_MEMORY;
    
    p_element->type = type;
    switch (type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        p_element->element.attribute = 
            ldap_str2attributetype( rawValue, &rc, &errorMessage, 
                                    LDAP_SCHEMA_ALLOW_ALL); 
        break;
    case LDAP_SCHEMA_OBJECT_CLASS:
        p_element->element.object = 
            ldap_str2objectclass( rawValue, &rc, &errorMessage, 
                                  LDAP_SCHEMA_ALLOW_ALL); 
        break;
    case LDAP_SCHEMA_SYNTAX:
        p_element->element.syntax = 
            ldap_str2syntax ( rawValue, &rc, &errorMessage, 
                              LDAP_SCHEMA_ALLOW_ALL); 
        break;
    case LDAP_SCHEMA_MATCHING_RULE:
        p_element->element.matching = 
            ldap_str2matchingrule ( rawValue, &rc, &errorMessage, 
                                    LDAP_SCHEMA_ALLOW_ALL); 
        break;
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        p_element->element.matchinguse = 
            ldap_str2matchingruleuse ( rawValue, &rc, &errorMessage, 
                                       LDAP_SCHEMA_ALLOW_ALL); 
        break;
    case LDAP_SCHEMA_NAME_FORM:
        p_element->element.nameform = 
            ldap_str2nameform ( rawValue, &rc, &errorMessage, 
                                LDAP_SCHEMA_ALLOW_ALL); 
        break;
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        p_element->element.ditcontent = 
            ldap_str2ditcontentrule( rawValue, &rc, &errorMessage, 
                                     LDAP_SCHEMA_ALLOW_ALL); 
        break;
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:
        p_element->element.ditstructure = 
            ldap_str2ditstructurerule( rawValue, &rc, &errorMessage, 
                                       LDAP_SCHEMA_ALLOW_ALL); 
        break;
    default:
        rc = LDAP_PARAM_ERROR;
    }

    if (rc == LDAP_SCHERR_OUTOFMEM) {
        rc = LDAP_NO_MEMORY;
    }
    if (rc != LDAP_SUCCESS) {
        LDAP_FREE( p_element );
        if (rc == LDAP_PARAM_ERROR || rc == LDAP_NO_MEMORY) {
            return rc;
        }
        else{
            return LDAP_LOCAL_ERROR;        
        }
    }

    rc = add_element(schema, p_element);
    return rc;
}



/*********************************************************************
    add_element:
    To be used by parse_element and ldap_schema_add & modify.
    Assumes the element has memory allocated and it's fields are valid. 
    This allocates more memory if needed and adds element to the oid
    list and names list. remove_element does the opposite.
 */
int add_element(
               LDAPSchema *schema, 
               LDAPSchemaElement *p_element)
{
    int i = 0, type = p_element->type;
    char** names = NULL;
    int count = 0;
    switch (type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        names = p_element->element.attribute->at_names;
        break;
    case LDAP_SCHEMA_OBJECT_CLASS:
        names = p_element->element.object->oc_names;
        break;
    case LDAP_SCHEMA_SYNTAX:
        names = p_element->element.syntax->syn_names;
        break;
    case LDAP_SCHEMA_MATCHING_RULE:
        names = p_element->element.matching->mr_names;
        break;
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        names = p_element->element.matchinguse->mru_names;
        break;
    case LDAP_SCHEMA_NAME_FORM:
        names = p_element->element.nameform->nf_names;
        break;
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        names = p_element->element.ditcontent->cr_names;
        break;
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:
        names = p_element->element.ditstructure->sr_names;
        break;
    default:
        assert(0);
        LDAP_FREE( p_element );
        return LDAP_PARAM_ERROR;
    }
    /* oid list */
    if (schema->oidCount[type] >= schema->oidsAllocated[type]) {
        schema->oidsAllocated[type] += 10;    
        /* 10 is arbitrary, could be any number */
        schema->oids[type] = 
            LDAP_REALLOC( schema->oids[type],
                          schema->oidsAllocated[type] * 
                          sizeof(LDAPSchemaElement* ));
        if ( schema->oids[type] == NULL ){
            LDAP_FREE( p_element );
            return LDAP_NO_MEMORY;
        }
            
    }
    schema->oids[type][schema->oidCount[type]] = p_element;
    schema->oidCount[type]++;
    count = ldap_count_values( names );

    /* name list */
    if ( (schema->nameCount[type] + count) >= schema->namesAllocated[type] ) {
        schema->namesAllocated[type] += count + 10; 
        /* we add 10 so we don't have to allocate on each add */
        schema->names[type] = 
            LDAP_REALLOC( schema->names[type],
                          schema->namesAllocated[type] * 
                          sizeof(struct ldap_schema_named_element));
        if ( schema->names[type] == NULL )
            return LDAP_NO_MEMORY;
    }
    /* Add an entry for each name */
    for (i=0; i<count; i++) {
        schema->names[type][schema->nameCount[type]].name = 
            LDAP_STRDUP(names[i]);
        if (schema->names[type][schema->nameCount[type]].name == NULL) {
            return LDAP_NO_MEMORY;
        }
        schema->names[type][schema->nameCount[type]].p_element = p_element;
        schema->nameCount[type] ++;
    }    
    return LDAP_SUCCESS;
}



/*  The following functions are used by Quick Sort and Binary Search  */

/*********************************************************************
    compare_oid:
    Used to compare elements by oid for qsort.
 */
int compare_oid( const void* element1, const void *element2){
    LDAPSchemaElement *a, *b;
    a = *((LDAPSchemaElement **)element1), 
    b = *((LDAPSchemaElement **)element2);

    if ( (a==NULL) && (b==NULL) )
        return 0;    
    if ( a == NULL)
        return 1;   /* greater than */
    if ( b == NULL)
        return -1;  /* less than */

    switch (a->type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        return strcmp ( a->element.attribute->at_oid,
                        b->element.attribute->at_oid);
    case LDAP_SCHEMA_OBJECT_CLASS:
        return strcmp ( a->element.object->oc_oid,
                        b->element.object->oc_oid );
    case LDAP_SCHEMA_SYNTAX:
        return strcmp ( a->element.syntax->syn_oid,
                        b->element.syntax->syn_oid );
    case LDAP_SCHEMA_MATCHING_RULE:
        return strcmp ( a->element.matching->mr_oid,
                        b->element.matching->mr_oid);
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        return strcmp ( a->element.matchinguse->mru_oid,
                        b->element.matchinguse->mru_oid);
    case LDAP_SCHEMA_NAME_FORM:
        return strcmp ( a->element.nameform->nf_oid,
                        b->element.nameform->nf_oid);
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        return strcmp ( a->element.ditcontent->cr_oid,
                        b->element.ditcontent->cr_oid);
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:
        return strcmp ( a->element.ditstructure->sr_rule_id,
                        b->element.ditstructure->sr_rule_id);
    default:
        assert (0);
        return a->type - b->type;
        /* this shouldn't happen, but I thought I'd sort by type anyway*/
    }
}


/*********************************************************************
    compare_name:
    used to compare two ldap_schema_named_elements by their name.
    Used by qsort.
 */
int compare_name( const void* element1, const void *element2){
    struct ldap_schema_named_element *a, *b;
    a = (struct ldap_schema_named_element *)element1;
    b = (struct ldap_schema_named_element *)element2;

    assert ( (a != NULL) && (b != NULL) );

    if ( (a->name == NULL) && (b->name == NULL) )
        return 0;    
    if ( a->name == NULL)
        return 1;   /* greater than */
    if ( b->name == NULL)
        return -1;  /* less than */
    
    return strcasecmp( a->name, b->name );
}


/*********************************************************************
    compare_key_to_oid:
    used to compare a key to an oid in a schema element. Used by bsearch.
 */
int compare_key_to_oid( const void* key, const void *schemaElement ){
    LDAPSchemaElement * element;
    char * oid;

    assert ( (schemaElement != NULL) && (key != NULL));
    
    element = *((LDAPSchemaElement **) schemaElement);

    oid = (char*)key;
    
    switch (element->type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        return strcmp ( oid, element->element.attribute->at_oid);
    case LDAP_SCHEMA_OBJECT_CLASS:
        return strcmp ( oid, element->element.object->oc_oid );
    case LDAP_SCHEMA_SYNTAX:
        return strcmp ( oid, element->element.syntax->syn_oid );
    case LDAP_SCHEMA_MATCHING_RULE:
        return strcmp ( oid, element->element.matching->mr_oid);
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        return strcmp ( oid, element->element.matchinguse->mru_oid);
    case LDAP_SCHEMA_NAME_FORM:
        return strcmp ( oid, element->element.nameform->nf_oid);
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        return strcmp ( oid, element->element.ditcontent->cr_oid);
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:
        return strcmp ( oid, element->element.ditstructure->sr_rule_id);
    default:
        assert(0);
        return 1;
    }
}


/*********************************************************************
    compare_key_to_name:
    Compares a key string to the name of a named element. Used by bsearch.
 */
int compare_key_to_name( const void* key, const void *namedElement){
    assert (key != NULL);
    if (((struct ldap_schema_named_element *)namedElement)->name == NULL)
        return -1;  /* lesser than */

    return strcasecmp( 
        (char*)key, 
        ((struct ldap_schema_named_element *)namedElement)->name );
}


/**********************************************************************
    ldap_schema_get_by_name 

    Retrieves a schema element from the list specified by elementType 
    according to the name passed in. 

    Uses a binary search - assumes the lists are sorted.
*/
LDAP_F( int )
ldap_schema_get_by_name(
                       LDAPSchema* schema, 
                       char* name_or_oid, 
                       int elementType, 
                       LDAPSchemaElement **element)
{
    struct ldap_schema_named_element *namedEl;
    LDAPSchemaElement **temp = NULL;

    *element = NULL;

    if (schema->testInit != KEYWORD)
        return LDAP_LOCAL_ERROR;

    if ((name_or_oid == NULL) || (schema == NULL)   ||
        (elementType < LDAP_SCHEMA_ATTRIBUTE_TYPE)  ||
        (elementType > LDAP_SCHEMA_DIT_STRUCTURE_RULE) )
        return LDAP_PARAM_ERROR;
    
    if (isdigit(name_or_oid[0])) {
        /* oid lookup */
        if ((schema->oids[elementType] == NULL) ||
            (schema->oidCount[elementType] == 0)) {
            return LDAP_NO_SUCH_OBJECT;
        }
        temp = (LDAPSchemaElement**)
                     bsearch((void*) name_or_oid,                /* key  */
                             (void*) schema->oids[elementType],  /* list */
                             schema->oidCount[elementType],                   
                             sizeof( LDAPSchemaElement *),
                             compare_key_to_oid  );
        if (temp == NULL)
            return LDAP_NO_SUCH_OBJECT;
        *element = *temp;

        if (*element == NULL)
            return LDAP_NO_SUCH_OBJECT;
    } else {
        /* name lookup */
        if ((schema->names[elementType] == NULL)  ||
            (schema->nameCount[elementType] == 0)) {
            return LDAP_NO_SUCH_OBJECT;
        }
        namedEl  = (struct ldap_schema_named_element *)
                   bsearch((void*) name_or_oid,                 /* key  */
                           (void*) schema->names[elementType],  /* list */
                           schema->nameCount[elementType],                   
                           sizeof( struct ldap_schema_named_element ),
                           compare_key_to_name );
        if (namedEl == NULL)
            return LDAP_NO_SUCH_OBJECT;
        *element = namedEl->p_element;        
    }   
    return LDAP_SUCCESS;
}



/**********************************************************************
     ldap_schema_get_by_index:
    Retrieve an element by index and type
 */
LDAP_F( int )
ldap_schema_get_by_index(
                        LDAPSchema* schema, 
                        int index, 
                        int elementType, 
                        LDAPSchemaElement **element)
{
    *element = NULL;
    if ((schema == NULL) || 
        (elementType < LDAP_SCHEMA_ATTRIBUTE_TYPE) ||
        (elementType > LDAP_SCHEMA_DIT_STRUCTURE_RULE)   ||
        (index >= schema->oidCount[elementType])   ||
        (index < 0))
        return LDAP_PARAM_ERROR;

    if (schema->testInit != KEYWORD)
        return LDAP_LOCAL_ERROR;

    *element = schema->oids[elementType][index];
    return LDAP_SUCCESS;
}



/**********************************************************************
    ldap_schema_get_count:
    retrieve the count of elements of a specific type.
 */
LDAP_F( int )
ldap_schema_get_count (LDAPSchema* schema, int elementType)
{
    if ((schema == NULL) || 
        (elementType < LDAP_SCHEMA_ATTRIBUTE_TYPE) ||
        (elementType > LDAP_SCHEMA_DIT_STRUCTURE_RULE))
        return -1;

    if (schema->testInit != KEYWORD)
        return -1;

    return schema->oidCount[elementType];
}



/********************************************************************** 
    ldap_schema_get_field_names:
    Generates a list of field names for an element
 */
LDAP_F( int )
ldap_schema_get_field_names(
                           LDAPSchemaElement *element, 
                           char *(*fields[]))
{
    LDAPSchemaExtensionItem **x_fields = NULL;     /*list of X- qualifiers   */
    int length = 0;                         /*1 for null;             */
    char *normal_fields[14];                /*13=Max number of fields */
    int temp = 0,x_count = 0,i = 0;

    switch (element->type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        if (element->element.attribute->at_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OID );
        if (element->element.attribute->at_names != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAMES );
        if (element->element.attribute->at_desc != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DESCRIPTION );
        if (element->element.attribute->at_sup_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_SUPERIOR );
        if (element->element.attribute->at_equality_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_EQUALITY );
        if (element->element.attribute->at_ordering_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_ORDERING );
        if (element->element.attribute->at_substr_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_SUBSTRING );
        //syntax len will be the second value of syntax
        if (element->element.attribute->at_syntax_oid != NULL) 
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_SYNTAX_OID );
        if (element->element.attribute->at_obsolete == 1)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OBSOLETE );
        if (element->element.attribute->at_single_value == 1)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_SINGLE_VALUED );
        if (element->element.attribute->at_collective == 1)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_COLLECTIVE );
        if (element->element.attribute->at_no_user_mod == 1)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NO_USER_MOD );

        /* ADD USAGE */
        temp = element->element.attribute->at_usage;
        if (temp == LDAP_SCHEMA_USER_APPLICATIONS)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_USER_APP );
        else if (temp == LDAP_SCHEMA_DIRECTORY_OPERATION)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DIRECTORY_OP );
        else if (temp == LDAP_SCHEMA_DISTRIBUTED_OPERATION)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DISTRIBUTED_OP );
        else if (temp == LDAP_SCHEMA_DSA_OPERATION)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DSA_OP );

        x_fields = element->element.attribute->at_extensions;
        break;
    case LDAP_SCHEMA_OBJECT_CLASS:
        if (element->element.object->oc_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OID );
        if (element->element.object->oc_names != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAMES );
        if (element->element.object->oc_desc != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DESCRIPTION );
        if (element->element.object->oc_sup_oids != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_SUPERIORS );
        if (element->element.object->oc_at_oids_must != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_MUST_ATTRIBUTES );
        if (element->element.object->oc_at_oids_may != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_MAY_ATTRIBUTES );
        if (element->element.object->oc_obsolete == 1)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OBSOLETE );

        /* add kind: abstract, structural auxiliary */
        temp = element->element.object->oc_kind;
        if ( temp == LDAP_SCHEMA_ABSTRACT)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_TYPE_ABSTRACT );
        else if ( temp == LDAP_SCHEMA_STRUCTURAL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_TYPE_STRUCTURAL );
        else if ( temp == LDAP_SCHEMA_AUXILIARY)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_TYPE_AUXILIARY );

        x_fields = element->element.object->oc_extensions;
        break;
    case LDAP_SCHEMA_SYNTAX:
        if ( element->element.syntax->syn_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OID );
        if ( element->element.syntax->syn_names != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAMES );
        if ( element->element.syntax->syn_desc != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DESCRIPTION );
        x_fields = element->element.syntax->syn_extensions;
        break;
    case LDAP_SCHEMA_MATCHING_RULE:
        if ( element->element.matching->mr_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OID );
        if ( element->element.matching->mr_names != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAMES );
        if ( element->element.matching->mr_desc != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DESCRIPTION );
        if ( element->element.matching->mr_syntax_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_SYNTAX_OID );
        if ( element->element.matching->mr_obsolete == 1)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OBSOLETE );
        x_fields = element->element.matching->mr_extensions;
        break;
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        if ( element->element.matchinguse->mru_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OID );
        if ( element->element.matchinguse->mru_desc!= NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DESCRIPTION );
        if ( element->element.matchinguse->mru_names != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAMES );
        if ( element->element.matchinguse->mru_at_applies != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_APPLIES );
        if ( element->element.matchinguse->mru_obsolete == 1)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OBSOLETE );
        x_fields = element->element.matchinguse->mru_extensions;
        break;
    case LDAP_SCHEMA_NAME_FORM:
        if ( element->element.nameform->nf_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OID );
        if ( element->element.nameform->nf_desc != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DESCRIPTION );
        if ( element->element.nameform->nf_names != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAMES );
        if ( element->element.nameform->nf_obsolete == 1 )
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OBSOLETE );
        if ( element->element.nameform->nf_oc_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAME_FORM_OBJECT );
        if ( element->element.nameform->nf_at_oids_may != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_MAY_ATTRIBUTES );
        if ( element->element.nameform->nf_at_oids_must != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_MUST_ATTRIBUTES );
        x_fields = element->element.nameform->nf_extensions;
        break;
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        if ( element->element.ditcontent->cr_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OID );
        if ( element->element.ditcontent->cr_desc != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DESCRIPTION );
        if ( element->element.ditcontent->cr_names != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAMES );
        if ( element->element.ditcontent->cr_obsolete == 1)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OBSOLETE );
        if ( element->element.ditcontent->cr_oids_aux != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_AUX_CLASSES );
        if ( element->element.ditcontent->cr_oids_must != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_MUST_ATTRIBUTES );
        if ( element->element.ditcontent->cr_oids_may != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_MAY_ATTRIBUTES );
        if ( element->element.ditcontent->cr_oids_not != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NOT_ATTRIBUTES );
        x_fields = element->element.ditcontent->cr_extensions;
        break;
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:            
        if ( element->element.ditstructure->sr_rule_id != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_RULE_ID );
        if ( element->element.ditstructure->sr_desc != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_DESCRIPTION );
        if ( element->element.ditstructure->sr_names != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAMES);
        if ( element->element.ditstructure->sr_obsolete == 1)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_OBSOLETE);
        if ( element->element.ditstructure->sr_nf_oid != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_NAME_FORM_OID);
        if ( element->element.ditstructure->sr_sup_ids != NULL)
            normal_fields[ length ++ ] = 
                LDAP_STRDUP( LDAP_SCHEMA_SUPERIOR);
        x_fields = element->element.ditstructure->sr_extensions;
        break;
    default:    
        assert(0);            
        return LDAP_PARAM_ERROR;
    }

    if (x_fields == NULL)
        x_count=0;
    else /* count x-qualifiers */
        for (x_count=0; x_fields[x_count] != NULL; x_count++)
            ;

    /* alloc for normal fields, qualifiers and terminating NULL */    
    (*fields) = LDAP_CALLOC( length + x_count + 1, sizeof(char*));

    /* copy in normal qualifiers */
    for (i=0; i<length; i++){
        (*fields)[i] = normal_fields[i];
        if (normal_fields[i] == NULL) {
            LDAP_FREE(*fields);
            return LDAP_NO_MEMORY;
        }
    }

    /* add qualifier names */
    if (x_fields != NULL){
        for (i=0; x_fields[i] != NULL; i++){
            (*fields)[length] = LDAP_STRDUP( x_fields[i]->lsei_name );
            if ((*fields)[length] == NULL){
                LDAP_FREE(*fields);
                return LDAP_NO_MEMORY;
            }
            length++;
        }
    }

    (*fields)[length] = NULL;
    return LDAP_SUCCESS;
}




/**********************************************************************
    ldap_schema_get_field_values:
    Generates a list of values of a specific field name, for an element.
 */
LDAP_F( int )
ldap_schema_get_field_values(
                            LDAPSchemaElement *element, 
                            char *name, 
                            char *(*values[]) )
{
    char maxSizeString[12]; /* an int may have 11 characters, -2147483647 */
    int temp = 0, x_count = 0;
    LDAPSchemaExtensionItem **x_fields = NULL;     //list of X- qualifiers   


    *values = NULL;
    switch (element->type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        if ( strcasecmp( name, LDAP_SCHEMA_OID ) == 0)
            return single_string_dup(
                values, element->element.attribute->at_oid);
        if ( strcasecmp( name, LDAP_SCHEMA_NAMES ) == 0)
            return double_string_dup(
                values, element->element.attribute->at_names);
        if ( strcasecmp( name, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return single_string_dup(
                values, element->element.attribute->at_desc);
        if ( strcasecmp( name, LDAP_SCHEMA_SUPERIOR ) == 0)
            return single_string_dup(
                values, element->element.attribute->at_sup_oid);
        if ( strcasecmp( name, LDAP_SCHEMA_EQUALITY ) == 0)
            return single_string_dup(
                values, element->element.attribute->at_equality_oid);
        if ( strcasecmp( name, LDAP_SCHEMA_ORDERING ) == 0)
            return single_string_dup(
                values, element->element.attribute->at_ordering_oid);
        if ( strcasecmp( name, LDAP_SCHEMA_SUBSTRING ) == 0)
            return single_string_dup(
                values, element->element.attribute->at_substr_oid);
        if ( strcasecmp( name, LDAP_SCHEMA_SYNTAX_OID ) == 0) {
            //the syntax length is the second value if it exists
            if (element->element.attribute->at_syntax_oid != NULL) {
                *values = LDAP_MALLOC( 3 * sizeof(char*));
                if (*values == NULL)
                    return LDAP_NO_MEMORY;

                (*values)[0] = LDAP_STRDUP( 
                    element->element.attribute->at_syntax_oid );
                if ((*values)[0] == NULL) 
                    return LDAP_NO_MEMORY;
                
                sprintf(maxSizeString, "%d", 
                        element->element.attribute->at_syntax_len );
                (*values)[1] = LDAP_STRDUP( maxSizeString );
                (*values)[2] = NULL;

                if ((*values)[1] == NULL) 
                    return LDAP_NO_MEMORY;
                return LDAP_SUCCESS;
            } else {
                values = NULL;
                return LDAP_NO_SUCH_ATTRIBUTE;
            }
        }
        if ( strcasecmp( name, LDAP_SCHEMA_OBSOLETE ) == 0) {
            if (element->element.attribute->at_obsolete)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        if ( strcasecmp( name, LDAP_SCHEMA_SINGLE_VALUED ) == 0) {
            if (element->element.attribute->at_single_value)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        if ( strcasecmp( name, LDAP_SCHEMA_COLLECTIVE ) == 0) {
            if (element->element.attribute->at_collective)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        if ( strcasecmp( name, LDAP_SCHEMA_NO_USER_MOD ) == 0) {
            if (element->element.attribute->at_no_user_mod)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }

        /* ADD USAGE */
        temp = element->element.attribute->at_usage;
        if ( strcasecmp( name, LDAP_SCHEMA_USER_APP ) == 0) {
            if (temp == LDAP_SCHEMA_USER_APPLICATIONS)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        if ( strcasecmp( name, LDAP_SCHEMA_DIRECTORY_OP ) == 0) {
            if (temp == LDAP_SCHEMA_DIRECTORY_OPERATION)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        } else if ( strcasecmp( name, LDAP_SCHEMA_DISTRIBUTED_OP ) == 0) {
            if (temp == LDAP_SCHEMA_DISTRIBUTED_OPERATION)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        } else if ( strcasecmp( name, LDAP_SCHEMA_DSA_OP ) == 0) {
            if (temp == LDAP_SCHEMA_DSA_OPERATION)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        x_fields = element->element.attribute->at_extensions;
        break;
    case LDAP_SCHEMA_OBJECT_CLASS:
        if ( strcasecmp( name, LDAP_SCHEMA_OID ) == 0)
            return single_string_dup(
                values, element->element.object->oc_oid);
        if ( strcasecmp( name, LDAP_SCHEMA_NAMES ) == 0)
            return double_string_dup(
                values, element->element.object->oc_names );
        if ( strcasecmp( name, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return single_string_dup(
                values, element->element.object->oc_desc );
        if ( strcasecmp( name, LDAP_SCHEMA_SUPERIORS ) == 0)
            return double_string_dup(
                values, element->element.object->oc_sup_oids );
        if ( strcasecmp( name, LDAP_SCHEMA_MUST_ATTRIBUTES ) == 0)
            return double_string_dup(
                values, element->element.object->oc_at_oids_must );
        if ( strcasecmp( name, LDAP_SCHEMA_MAY_ATTRIBUTES ) == 0)
            return double_string_dup(
                values, element->element.object->oc_at_oids_may );

        if ( strcasecmp( name, LDAP_SCHEMA_OBSOLETE ) == 0) {
            if ( element->element.object->oc_obsolete == 1 )
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }

        /* add kind: abstract, structural auxiliary */
        temp = element->element.object->oc_kind;
        if ( strcasecmp( name, LDAP_SCHEMA_TYPE_ABSTRACT ) == 0) {
            if (temp == LDAP_SCHEMA_ABSTRACT)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        } else if ( strcasecmp( name, LDAP_SCHEMA_TYPE_STRUCTURAL ) == 0) {
            if (temp == LDAP_SCHEMA_STRUCTURAL)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        } else if ( strcasecmp( name, LDAP_SCHEMA_TYPE_AUXILIARY  ) == 0) {
            if (temp == LDAP_SCHEMA_AUXILIARY)
                return single_string_dup(values, "");
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        x_fields = element->element.object->oc_extensions;
        break;
    case LDAP_SCHEMA_SYNTAX:
        if ( strcasecmp( name, LDAP_SCHEMA_OID ) == 0)
            return single_string_dup(
                values, element->element.syntax->syn_oid );
        if ( strcasecmp( name, LDAP_SCHEMA_NAMES ) == 0)
            return double_string_dup(
                values, element->element.syntax->syn_names );
        if ( strcasecmp( name, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return single_string_dup(
                values, element->element.syntax->syn_desc );
        x_fields = element->element.syntax->syn_extensions;
        break;
    case LDAP_SCHEMA_MATCHING_RULE:
        if ( strcasecmp( name, LDAP_SCHEMA_OID ) == 0)
            return single_string_dup(
                values, element->element.matching->mr_oid );
        if ( strcasecmp( name, LDAP_SCHEMA_NAMES ) == 0)
            return double_string_dup(
                values, element->element.matching->mr_names );
        if ( strcasecmp( name, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return single_string_dup(
                values, element->element.matching->mr_desc );
        if ( strcasecmp( name, LDAP_SCHEMA_SYNTAX_OID ) == 0)
            return single_string_dup(
                values, element->element.matching->mr_syntax_oid );
        if ( strcasecmp( name, LDAP_SCHEMA_OBSOLETE ) == 0) {
            if (element->element.matching->mr_obsolete != 0)
                return single_string_dup(values, "" );
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        x_fields = element->element.matching->mr_extensions;
        break;
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        if ( strcasecmp( name, LDAP_SCHEMA_OID ) == 0)
            return single_string_dup(
                values, element->element.matchinguse->mru_oid );
        if ( strcasecmp( name, LDAP_SCHEMA_NAMES ) == 0)
            return double_string_dup(
                values, element->element.matchinguse->mru_names);
        if ( strcasecmp( name, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return single_string_dup(
                values, element->element.matchinguse->mru_desc);
        if ( strcasecmp( name, LDAP_SCHEMA_OBSOLETE ) == 0) {
            if (element->element.matchinguse->mru_obsolete != 0)
                return single_string_dup(values, "" );
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        if ( strcasecmp( name, LDAP_SCHEMA_APPLIES ) == 0)
            return double_string_dup(
                values, element->element.matchinguse->mru_at_applies);
        x_fields = element->element.matchinguse->mru_extensions;
        break;
    case LDAP_SCHEMA_NAME_FORM:
        if ( strcasecmp( name, LDAP_SCHEMA_OID ) == 0)
            return single_string_dup(
                values, element->element.nameform->nf_oid );
        if ( strcasecmp( name, LDAP_SCHEMA_NAMES ) == 0)
            return double_string_dup(
                values, element->element.nameform->nf_names);
        if ( strcasecmp( name, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return single_string_dup(
                values, element->element.nameform->nf_desc);
        if ( strcasecmp( name, LDAP_SCHEMA_OBSOLETE ) == 0) {
            if (element->element.nameform->nf_obsolete != 0)
                return single_string_dup(values, "" );
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        if ( strcasecmp( name, LDAP_SCHEMA_NAME_FORM_OBJECT ) == 0)
            return single_string_dup(
                values, element->element.nameform->nf_oc_oid);
        if ( strcasecmp( name, LDAP_SCHEMA_MUST_ATTRIBUTES ) == 0)
            return double_string_dup(
                values, element->element.nameform->nf_at_oids_must);
        if ( strcasecmp( name, LDAP_SCHEMA_MAY_ATTRIBUTES ) == 0)
            return double_string_dup(
                values, element->element.nameform->nf_at_oids_may);
        x_fields = element->element.matchinguse->mru_extensions;
        break;
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        if ( strcasecmp( name, LDAP_SCHEMA_OID ) == 0)
            return single_string_dup(
                values, element->element.ditcontent->cr_oid );
        if ( strcasecmp( name, LDAP_SCHEMA_NAMES ) == 0)
            return double_string_dup(
                values, element->element.ditcontent->cr_names);
        if ( strcasecmp( name, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return single_string_dup(
                values, element->element.ditcontent->cr_desc);
        if ( strcasecmp( name, LDAP_SCHEMA_OBSOLETE ) == 0) {
            if (element->element.ditcontent->cr_obsolete != 0)
                return single_string_dup(values, "" );
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        if ( strcasecmp( name, LDAP_SCHEMA_AUX_CLASSES ) == 0)
            return double_string_dup(
                values, element->element.ditcontent->cr_oids_aux);
        if ( strcasecmp( name, LDAP_SCHEMA_MUST_ATTRIBUTES ) == 0)
            return double_string_dup(
                values, element->element.ditcontent->cr_oids_must);
        if ( strcasecmp( name, LDAP_SCHEMA_MAY_ATTRIBUTES ) == 0)
            return double_string_dup(
                values, element->element.ditcontent->cr_oids_may);
        if ( strcasecmp( name, LDAP_SCHEMA_NOT_ATTRIBUTES ) == 0)
            return double_string_dup(
                values, element->element.ditcontent->cr_oids_not);

        x_fields = element->element.matchinguse->mru_extensions;
        break;
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:            
        if ( strcasecmp( name, LDAP_SCHEMA_RULE_ID ) == 0)
            return single_string_dup(
                values, element->element.ditstructure->sr_rule_id );
        if ( strcasecmp( name, LDAP_SCHEMA_NAMES ) == 0)
            return double_string_dup(
                values, element->element.ditstructure->sr_names );
        if ( strcasecmp( name, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return single_string_dup(
                values, element->element.ditstructure->sr_desc);
        if ( strcasecmp( name, LDAP_SCHEMA_OBSOLETE ) == 0) {
            if (element->element.ditstructure->sr_obsolete != 0)
                return single_string_dup(values, "" );
            else
                return LDAP_NO_SUCH_ATTRIBUTE;
        }
        if ( strcasecmp( name, LDAP_SCHEMA_NAME_FORM_OID ) == 0)
            return single_string_dup(
                values, element->element.ditstructure->sr_nf_oid );
        if ( strcasecmp( name, LDAP_SCHEMA_SUPERIOR ) == 0)
            return double_string_dup(
                values, element->element.ditstructure->sr_sup_ids );
             x_fields = element->element.ditstructure->sr_extensions;
        break;
    default:                
        assert(0);
        return LDAP_PARAM_ERROR;
    }

    if (x_fields == NULL)
        return LDAP_NO_SUCH_ATTRIBUTE;
    
    /*searching x_fields for the qualifier in 'name' */
    x_count=0; 
    while ( (x_fields[x_count] != NULL) && 
            (strcasecmp( name, x_fields[x_count]->lsei_name ) != 0) )
        x_count++;
    if (x_fields[x_count] == NULL)
        return LDAP_NO_SUCH_ATTRIBUTE;

    return double_string_dup(values, x_fields[x_count]->lsei_values);  
}


/*********************************************************************
    str_array_dup:
    Duplicates an array of char* 
 */
char ** str_array_dup( char *stringArray[] )
{
    int i = 0, length = 0;
    char ** newArray = NULL;
    if (stringArray == NULL)
        return NULL;
    length = ldap_count_values( stringArray );
    newArray = LDAP_CALLOC( length + 1, sizeof(char*));    
    for (i = 0; i<length; i++){
        if (stringArray[i] == NULL) {
            newArray[i] = NULL;
        }else{
            newArray[i] = LDAP_STRDUP( stringArray[i] );
            if (newArray[i] == NULL) {
                LDAP_VFREE( newArray );
                return NULL;
            }   
        }
    }
    newArray[i] = NULL;
    return newArray;
}



/*********************************************************************
    single_string_dup:
    Used exclusively by ldap_get_field_values.  Checks the field for
    validity and returns appropriate error code.  If valid the single
    string is copied into a char* array. 
 */
int single_string_dup( char *(*newValue[]), char* toDup)
{
    *newValue = NULL;
    if (toDup == NULL)
        return LDAP_NO_SUCH_ATTRIBUTE;
    else {
        *newValue = LDAP_CALLOC( 2, sizeof(char*));
        if (*newValue == NULL) 
            return LDAP_NO_MEMORY;

        (*newValue)[0] = LDAP_STRDUP(toDup);
        (*newValue)[1] = NULL;

        if ((*newValue)[0]== NULL) {
            LDAP_FREE (*newValue);
            return LDAP_NO_MEMORY;
        }
        return LDAP_SUCCESS;
    }    
}
 

/*********************************************************************
    double_string_dup:
    Used Exclusively by ldap_get_field_values.  Checks the field for
    validity and returns appropriate error code.  If valid the array of 
    strings is copied into a char* array. 
 */
int double_string_dup( char *(*newValue[]), char *toDup[])
{
    *newValue = NULL;
    if (toDup == NULL)
        return LDAP_NO_SUCH_ATTRIBUTE;
    else {
        (*newValue) = str_array_dup( toDup );
        return LDAP_SUCCESS;
    }
}



/********************************************************************** 	
    ldap_schema_delete:
    Deletes the schema element definition identified by the string
    name_oid.  The deletion operation will be stored in 'changes'.

    Code flow:
    1) Locate schema element.
    2) Store delete in the list of changes.
    3) Delete from the schema structure.
    
 */

LDAP_F( int )
ldap_schema_delete(
                  LDAPSchema *schema, 
                  int type,
                  char *name_oid )
{    
    LDAPSchemaElement  *toDelete = NULL;
    int rc = LDAP_SUCCESS;

    if ((schema == NULL) || (name_oid == NULL) ||  
        (type < LDAP_SCHEMA_ATTRIBUTE_TYPE) ||
        (type > LDAP_SCHEMA_DIT_STRUCTURE_RULE))
        return LDAP_PARAM_ERROR;


    /* 1) Locate schema element. */
    rc = ldap_schema_get_by_name(schema, name_oid, type, &toDelete);

    if (rc != LDAP_SUCCESS)
        return rc;

    /* 2) Store this deletion */
    rc = store_update(schema, toDelete, LDAP_MOD_DELETE);

    if (rc != LDAP_SUCCESS)
        return rc;

    /* 3) Delete from the schema structure. remove_element can only return
          LDAP_NO_SUCH_OBJECT.  Since we already checked this we can ignore
          the return value.
     */
    rc = remove_element(schema, toDelete->element.object->oc_oid, type);
 
    return rc;
}



/**********************************************************************
    ldap_schema_modify:
    Modifies the fields of a schema element.  Each mod structure in 
    the fieldsToChange array will have a flag to specify a delete, 
    replace or add.  The schema element is identified by name_oid.
    If the schema element is NULL then a temporary one is created.

    Code flow:
    1) Locate schema element
    2) Copy element for backup
    3) Modify element (in 'schema')
    4) Store modification in 'changes'
    5) Backout if needed
    6) Re-sort if
        1- backout occured
        2- oid changed
        3- name changed
 */
LDAP_F ( int )
ldap_schema_modify(
                  LDAPSchema *schema, 
                  char* name_oid,
                  int type, 
                  LDAPSchemaMod   *fields[])
{    
    LDAPSchemaElement *element = NULL, *backup = NULL;
    struct ldapmod del_el, add_el;
    char *del_val[2], *add_val[2], *typeString = NULL, *old_val = NULL,*new_val = NULL; 
    int rc = 0, i = 0, oid_changed = 0, names_changed = 0, backout = 0;

    old_val = NULL;
    new_val = NULL;

    if ((schema == NULL) || (name_oid == NULL) || (fields == NULL) || 
        (type < LDAP_SCHEMA_ATTRIBUTE_TYPE) ||
        (type > LDAP_SCHEMA_DIT_STRUCTURE_RULE))
        return LDAP_PARAM_ERROR;

    /* 1) Locate schema element */
    rc = ldap_schema_get_by_name(schema, name_oid, type, &element);
    if (rc != LDAP_SUCCESS)
        return rc;

    /* 2) Copy element for backup */
    rc = dup_element( element, &backup );     
    if (rc != LDAP_SUCCESS)
        return rc;

    /* 3) Modify element */
    for (i=0; fields[i] != NULL; i++ ) {
        /* Every thing after modify_element must be backed out
           in case of an error. */        
        rc = modify_element( element, fields[i]->fieldName, 
                        fields[i]->values, fields[i]->op );
        if (rc != LDAP_SUCCESS)
            goto BACKOUT;
        if (strcasecmp(fields[i]->fieldName, LDAP_SCHEMA_OID) == 0)
            oid_changed = 1;
        if (strcasecmp(fields[i]->fieldName, LDAP_SCHEMA_NAMES) == 0)
            names_changed = 1;
    }

    /* 4) add the Update to list of changes */
    rc = ldap_schema_element_to_string( backup, &old_val, &typeString); 
    if (rc != LDAP_SUCCESS) goto BACKOUT;
    del_val[0] = old_val;
    del_val[1] = NULL;
    del_el.mod_op = LDAP_MOD_DELETE;
    del_el.mod_values = del_val;
    del_el.mod_type = typeString;

    rc = ldap_schema_element_to_string( element, &new_val, &typeString ); 
    if (rc != LDAP_SUCCESS) goto BACKOUT;
    add_val[0] = new_val;
    add_val[1] = NULL;
    add_el.mod_op = LDAP_MOD_ADD;
    add_el.mod_values = add_val;
    add_el.mod_type = typeString;

    /* save the modification, 
        add change will duplicate any important info */
    rc = add_change( schema, &del_el );
    if (rc != LDAP_SUCCESS)
        goto BACKOUT;

    rc = add_change( schema, &add_el );

    if (rc != LDAP_SUCCESS) {

BACKOUT:    /* 5) Backout if needed */
        
        backout = 1;

        /* remove changes made to element and replace with a copy of the old */
        if (remove_element(schema, element->element.attribute->at_oid, type)
            != LDAP_SUCCESS) 
            assert (0);
   
        if (( schema != NULL) &&  /* only add if schema is not temporary */
            (add_element(schema, backup) != LDAP_SUCCESS))
            rc = LDAP_LOCAL_ERROR;     
    }
    else free_element(backup);
        
    
    /* 6 Re-sort */
    if (backout || oid_changed || names_changed) {
        /* re-sort */
        qsort(  (void*)schema->oids[type],         /* list to sort         */
                schema->oidCount[type],         /* number of elements   */
                sizeof( LDAPSchemaElement*),    /* size of each item    */
                compare_oid  );                 /* Comparison function  */
        qsort(  (void*)schema->names[type],
                schema->nameCount[type],
                sizeof( struct ldap_schema_named_element ),
                compare_name );       
    }
    if (old_val != NULL) LDAP_FREE( old_val );
    if (new_val != NULL) LDAP_FREE( new_val );

    return rc;
}


/**********************************************************************	
    ldap_schema_add:
    Adds a new schema element definition to the local copy of the schema.  

    Code Flow:
    1) Create a new element.
    2) Add fields to the element.
    3) Add to local copy
    4) Store addition to changes.
    5) Remove addition if there is no memory to store.
    6) Re-sort.
 */
LDAP_F( int )
ldap_schema_add(
               LDAPSchema *schema, 
               int type, 
               LDAPSchemaMod *fields[])
{    
    LDAPSchemaElement *element = NULL; 
    int i = 0 ,rc = 0;

    if ((schema == NULL) || (fields == NULL) || 
        (type < LDAP_SCHEMA_ATTRIBUTE_TYPE) ||
        (type > LDAP_SCHEMA_DIT_STRUCTURE_RULE))
        return LDAP_PARAM_ERROR;

    /* 1 Create a new element */    
    element = LDAP_CALLOC( 1, sizeof(LDAPSchemaElement));
    if (element == NULL)
        return LDAP_NO_MEMORY;
    element->type = type;

    switch (type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        element->element.attribute = LDAP_CALLOC( 1, sizeof(LDAPAttributeType) );
        if (element->element.attribute == NULL)
            rc = LDAP_NO_MEMORY;
        break;
    case LDAP_SCHEMA_OBJECT_CLASS:
        element->element.object = LDAP_CALLOC( 1, sizeof(LDAPObjectClass) );
        if (element->element.object == NULL)
            rc = LDAP_NO_MEMORY;
        break;            
    case LDAP_SCHEMA_SYNTAX:
        element->element.syntax = LDAP_CALLOC( 1, sizeof(LDAPSyntax) );
        if (element->element.syntax == NULL)
            rc = LDAP_NO_MEMORY;
        break;
    case LDAP_SCHEMA_MATCHING_RULE:
        element->element.matching = LDAP_CALLOC( 1, sizeof(LDAPMatchingRule) );
        if (element->element.matching == NULL)
            rc = LDAP_NO_MEMORY;
        break;
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        element->element.matchinguse = 
            LDAP_CALLOC( 1, sizeof(LDAPMatchingRuleUse));
        if (element->element.matchinguse == NULL)
            rc = LDAP_NO_MEMORY;
        break;            
    case LDAP_SCHEMA_NAME_FORM:
        element->element.nameform = LDAP_CALLOC( 1, sizeof(LDAPNameForm) );
        if (element->element.nameform == NULL)
            rc = LDAP_NO_MEMORY;
        break;            
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        element->element.ditcontent = LDAP_MALLOC(sizeof(LDAPDITContentRule));
        if (element->element.ditcontent == NULL)
            rc = LDAP_NO_MEMORY;
        break;            
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:
        element->element.ditstructure = 
            LDAP_CALLOC( 1, sizeof(LDAPDITStructureRule));
        if (element->element.ditstructure == NULL)
            rc = LDAP_NO_MEMORY;
        break;            
    }

    if (rc == LDAP_NO_MEMORY)
        goto CLEANUP;

    /* 2) Add fields to the element */
    for ( i=0; fields[i]!= NULL; i++ ){
        rc = modify_element( element, fields[i]->fieldName, 
                        fields[i]->values, LDAP_MOD_ADD);
        if (rc != LDAP_SUCCESS)
            goto CLEANUP;
    }

    /* 3) add to local copy */
    
    rc = add_element(schema, element);
    if (rc != LDAP_SUCCESS)
        goto CLEANUP;
    
    /* 4) Store addition to changes */
    rc = store_update(schema, element, LDAP_MOD_ADD);

    /* 5) Remove addition if there is no memory to store the change */
    if (rc != LDAP_SUCCESS){
        remove_element(
            schema, 
            element->element.attribute->at_oid, 
            element->type);             
    }
 
    /* 6) resort */
    qsort(  (void*)schema->oids[type],        /* list to sort        */
            schema->oidCount[type],           /* number of elements  */
            sizeof( LDAPSchemaElement*),      /* size of each item   */
            compare_oid  );                   /* Comparison function */
    qsort(  (void*)schema->names[type],
            schema->nameCount[type],
            sizeof( struct ldap_schema_named_element ),
            compare_name );
    return rc;    

CLEANUP:

    if (element != NULL)
        free_element(element);
    return rc;    

}



/*********************************************************************
    ldap_schema_element_to_string:
    Used to get the string and type string for ldap_schema_add, 
    modify, and delete.  This is not exposed. 

    Note: str must be freed using LDAP_FREE.  typeString must not be
    freed because they are #defines.
 */
int ldap_schema_element_to_string(
                                 LDAPSchemaElement* element, 
                                 char** str, 
                                 char**typeString)
{

    switch (element->type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        *str = ldap_attributetype2str(element->element.attribute);            
        *typeString= ATTRIBUTE_TYPES;
        break;
    case LDAP_SCHEMA_OBJECT_CLASS:
        *str = ldap_objectclass2str(element->element.object);
        *typeString= OBJECT_CLASSES;
        break;
    case LDAP_SCHEMA_SYNTAX:
        *str = ldap_syntax2str(element->element.syntax);
        *typeString= LDAP_SYNTAXES;
        break;
    case LDAP_SCHEMA_MATCHING_RULE:
        *str = ldap_matchingrule2str(element->element.matching);
        *typeString= MATCHING_RULES;
        break;
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        *str = ldap_matchingruleuse2str(element->element.matchinguse ); 
        *typeString= MATCHING_RULE_USE;
        break;
    case LDAP_SCHEMA_NAME_FORM:
        *str = ldap_nameform2str(element->element.nameform ); 
        *typeString= NAME_FORMS;
        break;
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        *str = ldap_ditcontentrule2str(element->element.ditcontent ); 
        *typeString= DIT_CONTENT_RULES;
        break;
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:
        *str = ldap_ditstructurerule2str(element->element.ditstructure ); 
        *typeString= DIT_STRUCTURE_RULES;
        break;
    default:
        assert (0);
        return LDAP_PARAM_ERROR;
    }

    if (*str == NULL)
        return LDAP_LOCAL_ERROR;
    /*could be out of memory or a problem generating the string */
    else
        return LDAP_SUCCESS;
}


/*********************************************************************
    store_update:
    Generates a string definition for an element and stores Mod structure. 
    This is used by add and delete.  Modify will create the Mod structure 
    itself.
 */
int store_update(LDAPSchema *schema,
              LDAPSchemaElement * el,
              int modop)
{    
    struct ldapmod mod, *modify[2];
    int rc = LDAP_SUCCESS;
    char *rawValue = NULL, *typeString = NULL;
    char *modValues[2];

    rc = ldap_schema_element_to_string( el, &rawValue, &typeString);
    if (rc != LDAP_SUCCESS)
        return rc;

    modValues[0]    = rawValue;
    modValues[1]    = NULL;
    mod.mod_op      = modop; 
    mod.mod_values  = modValues;
    mod.mod_type    = typeString;

    modify[0] = &mod;
    modify[1] = NULL;

    /* Save the change */
    rc = add_change(schema, &mod);

    if (rawValue != NULL) 
        LDAP_FREE(rawValue);

    return rc;
}




/*********************************************************************
    dup_element:
    Duplicates an element and it's contents. Used by Modify.
 */
int dup_element(
                LDAPSchemaElement *element, 
                LDAPSchemaElement **newElement)
{
    char *temp = NULL;
    const char *error = NULL;
    int rc = 0;

    *newElement= LDAP_CALLOC( 1, sizeof(LDAPSchemaElement));
    if (*newElement == NULL)
        return LDAP_NO_MEMORY;

    switch (element->type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        temp = ldap_attributetype2str(element->element.attribute);
        (*newElement)->element.attribute = 
        ldap_str2attributetype(temp, &rc, &error, LDAP_SCHEMA_ALLOW_ALL);
        break;
    case LDAP_SCHEMA_OBJECT_CLASS:
        temp = ldap_objectclass2str(element->element.object);
        (*newElement)->element.object = 
        ldap_str2objectclass (temp, &rc, &error, LDAP_SCHEMA_ALLOW_ALL);
        break;
    case LDAP_SCHEMA_SYNTAX:
        temp = ldap_syntax2str(element->element.syntax);
        (*newElement)->element.syntax = 
        ldap_str2syntax(temp, &rc, &error, LDAP_SCHEMA_ALLOW_ALL);
        break;
    case LDAP_SCHEMA_MATCHING_RULE:
        temp = ldap_matchingrule2str(element->element.matching);
        (*newElement)->element.matching = 
        ldap_str2matchingrule(temp, &rc, &error, LDAP_SCHEMA_ALLOW_ALL);
        break;
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        temp = ldap_matchingruleuse2str(element->element.matchinguse);
        (*newElement)->element.matchinguse = 
        ldap_str2matchingruleuse(temp, &rc, &error, LDAP_SCHEMA_ALLOW_ALL);
        break;
    case LDAP_SCHEMA_NAME_FORM:
        temp = ldap_nameform2str(element->element.nameform);
        (*newElement)->element.nameform = 
        ldap_str2nameform(temp, &rc, &error, LDAP_SCHEMA_ALLOW_ALL);
        break;
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        temp = ldap_ditcontentrule2str(element->element.ditcontent);
        (*newElement)->element.ditcontent = 
        ldap_str2ditcontentrule(temp, &rc, &error, LDAP_SCHEMA_ALLOW_ALL);
        break;
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:
        temp = ldap_ditstructurerule2str(element->element.ditstructure);
        (*newElement)->element.ditstructure= 
        ldap_str2ditstructurerule(temp, &rc, &error, LDAP_SCHEMA_ALLOW_ALL);
        break;
    default:
        assert(0);
        return LDAP_PARAM_ERROR;
    }
    LDAP_FREE(temp);
    (*newElement)->type = element->type;
    return LDAP_SUCCESS;
}



/*********************************************************************
    modify_element:
    Changes the values for the given field name.  If mod type is delete 
    the value is removed or, if it is the last value, set to NULL. 
    Replace deletes the existing values and replaces them. Add adds to
    the values. if the element passed in is new, it must have all 
    fields set to null and zero.
 */
int modify_element(
                  LDAPSchemaElement *element, 
                  char *fieldName, 
                  char *values[], 
                  int modType)
{
    LDAPSchemaExtensionItem ***p_xterms = NULL, *p_xterm = NULL, **temp = NULL; 
    int i=0,newSize = 0; /* used to resize extension lists */

    switch (element->type) {
    /* new_value() is used for single-valued fields.
       new_values() is used for multi-valued fields.
       no-value fields are handled individually.  */
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        if (strcasecmp(fieldName, LDAP_SCHEMA_OID) == 0)
            return new_value(&element->element.attribute->at_oid, 
                             values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_DESCRIPTION) == 0)
            return new_value(&element->element.attribute->at_desc, 
                             values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAMES ) == 0)
            return new_values(&element->element.attribute->at_names, 
                              values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_OBSOLETE ) == 0) {
            /* No value */
            element->element.attribute->at_obsolete = 
                (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        } else if (strcasecmp(fieldName, LDAP_SCHEMA_SUPERIOR ) == 0)
            return new_value(&element->element.attribute->at_sup_oid, 
                             values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_EQUALITY ) == 0)
            return new_value(&element->element.attribute->at_equality_oid, 
                             values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_ORDERING ) == 0)
            return new_value(&element->element.attribute->at_ordering_oid, 
                             values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_SUBSTRING ) == 0)
            return new_value(&element->element.attribute->at_substr_oid, 
                             values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_SYNTAX_OID ) == 0)
            return new_value(&element->element.attribute->at_syntax_oid, 
                             values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_SINGLE_VALUED ) == 0) {
            /* no value */
            element->element.attribute->at_single_value = 
                (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        } else if (strcasecmp(fieldName, LDAP_SCHEMA_COLLECTIVE ) == 0) {
            element->element.attribute->at_collective = 
                (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        } else if (strcasecmp(fieldName, LDAP_SCHEMA_NO_USER_MOD ) == 0) {
            element->element.attribute->at_no_user_mod = 
                (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        } else if (strcasecmp(fieldName, LDAP_SCHEMA_USAGE) == 0) {
            if ((modType == LDAP_MOD_DELETE) || (values == NULL) || 
                (values[0] == NULL))
            {
                /* user app is default */
                element->element.attribute->at_usage = 
                    LDAP_SCHEMA_USER_APPLICATIONS;
                return LDAP_SUCCESS;
            }
            else {    /* these are the same for replace and add */                                
                if (strcasecmp(values[0], LDAP_SCHEMA_USER_APP ) == 0) {
                    element->element.attribute->at_usage = 
                        LDAP_SCHEMA_USER_APPLICATIONS;
                    return LDAP_SUCCESS;
                } else if (strcasecmp(values[0], 
                                     LDAP_SCHEMA_DIRECTORY_OP) == 0) {
                    element->element.attribute->at_usage = 
                        LDAP_SCHEMA_DIRECTORY_OPERATION;
                    return LDAP_SUCCESS;
                } else if (strcasecmp(values[0], 
                                     LDAP_SCHEMA_DISTRIBUTED_OP) == 0) {
                    element->element.attribute->at_usage = 
                        LDAP_SCHEMA_DISTRIBUTED_OPERATION;
                    return LDAP_SUCCESS;
                } else if (strcasecmp(values[0], 
                                     LDAP_SCHEMA_DSA_OP ) == 0) {
                    element->element.attribute->at_usage = 
                        LDAP_SCHEMA_DSA_OPERATION;
                    return LDAP_SUCCESS;
                } else
                    return LDAP_PARAM_ERROR;
            }
        }
        p_xterms = &element->element.attribute->at_extensions;
        break;

    case LDAP_SCHEMA_OBJECT_CLASS:
        if (strcasecmp(fieldName, LDAP_SCHEMA_OID ) == 0)
            return new_value(&element->element.object->oc_oid, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return new_value(&element->element.object->oc_desc, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAMES ) == 0)
            return new_values(&element->element.object->oc_names, 
                              values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_OBSOLETE ) == 0) {
            /* boolean, no value */
            element->element.object->oc_obsolete = 
            (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        } else if (strcasecmp(fieldName, LDAP_SCHEMA_SUPERIORS ) == 0)
            return new_values(&element->element.object->oc_sup_oids, 
                              values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_MUST_ATTRIBUTES ) == 0)
            return new_values(&element->element.object->oc_at_oids_must, 
                              values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_MAY_ATTRIBUTES ) == 0)
            return new_values(&element->element.object->oc_at_oids_may, 
                              values, modType );
        else if (strcasecmp(fieldName, LDAP_SCHEMA_TYPE_ABSTRACT ) == 0) {
            if (modType == LDAP_MOD_DELETE)
                element->element.object->oc_kind = -1; /* nothing */
            else
                element->element.object->oc_kind = LDAP_SCHEMA_ABSTRACT;
            return LDAP_SUCCESS;
        } else if (strcasecmp(fieldName, LDAP_SCHEMA_TYPE_STRUCTURAL ) == 0) {
            if (modType == LDAP_MOD_DELETE)
                element->element.object->oc_kind = -1;
            else
                element->element.object->oc_kind = LDAP_SCHEMA_STRUCTURAL;
            return LDAP_SUCCESS;
        } else if (strcasecmp(fieldName, LDAP_SCHEMA_TYPE_AUXILIARY ) == 0) {
            if (modType == LDAP_MOD_DELETE)
                element->element.object->oc_kind = -1;
            else
                element->element.object->oc_kind = LDAP_SCHEMA_AUXILIARY;
            return LDAP_SUCCESS;
        }
        p_xterms = &element->element.object->oc_extensions;
        break;

    case LDAP_SCHEMA_SYNTAX:
        if (strcasecmp(fieldName, LDAP_SCHEMA_OID) == 0)
            return new_value(&element->element.syntax->syn_oid, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_DESCRIPTION) == 0)
            return new_value(&element->element.syntax->syn_desc, 
                             values, modType);
        /* Though rfc2252 doesn't specify a name for syntax OpenLDAP uses it */
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAMES ) == 0)
            return new_values(&element->element.syntax->syn_names, 
                              values, modType );
        p_xterms = &element->element.syntax->syn_extensions;
        break;

    case LDAP_SCHEMA_MATCHING_RULE:
        if (strcasecmp(fieldName, LDAP_SCHEMA_OID ) == 0)
            return new_value(&element->element.matching->mr_oid, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return new_value(&element->element.matching->mr_desc, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAMES ) == 0)
            return new_values(&element->element.matching->mr_names, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_OBSOLETE ) == 0) {
            /* no value */                
            element->element.matching->mr_obsolete = 
                (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        }
        else if (strcasecmp(fieldName, LDAP_SCHEMA_SYNTAX_OID ) == 0)
            return new_value(&element->element.matching->mr_syntax_oid, 
                             values, modType);
        p_xterms = &element->element.matching->mr_extensions;
        break;

    case LDAP_SCHEMA_MATCHING_RULE_USE:            
        if (strcasecmp(fieldName, LDAP_SCHEMA_OID ) == 0)
            return new_value(&element->element.matchinguse->mru_oid, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return new_value(&element->element.matchinguse->mru_desc, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAMES ) == 0)
            return new_values(&element->element.matchinguse->mru_names, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_OBSOLETE ) == 0) {
            /* no value */                
            element->element.matchinguse->mru_obsolete = 
                (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        }
        else if (strcasecmp(fieldName, LDAP_SCHEMA_APPLIES ) == 0)
            return new_values(&element->element.matchinguse->mru_at_applies, 
                              values, modType);

        p_xterms = &element->element.matchinguse->mru_extensions;
        break;
    case LDAP_SCHEMA_NAME_FORM:            
        if (strcasecmp(fieldName, LDAP_SCHEMA_OID ) == 0)
            return new_value(&element->element.nameform->nf_oid, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return new_value(&element->element.nameform->nf_desc, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAMES ) == 0)
            return new_values(&element->element.nameform->nf_names, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_OBSOLETE ) == 0) {
            /* no value */                
            element->element.nameform->nf_obsolete = 
                (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        }
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAME_FORM_OBJECT ) == 0)
            return new_value(&element->element.nameform->nf_oc_oid, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_MUST_ATTRIBUTES ) == 0)
            return new_values(&element->element.nameform->nf_at_oids_must, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_MAY_ATTRIBUTES ) == 0)
            return new_values(&element->element.nameform->nf_at_oids_may, 
                              values, modType);
        p_xterms = &element->element.nameform->nf_extensions;
        break;
    case LDAP_SCHEMA_DIT_CONTENT_RULE:            
        if (strcasecmp(fieldName, LDAP_SCHEMA_OID ) == 0)
            return new_value(&element->element.ditcontent->cr_oid, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return new_value(&element->element.ditcontent->cr_desc, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAMES ) == 0)
            return new_values(&element->element.ditcontent->cr_names, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_OBSOLETE ) == 0) {
            /* no value */                
            element->element.ditcontent->cr_obsolete = 
            (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        }
        else if (strcasecmp(fieldName, LDAP_SCHEMA_AUX_CLASSES ) == 0)
            return new_values(&element->element.ditcontent->cr_oids_aux, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_MUST_ATTRIBUTES ) == 0)
            return new_values(&element->element.ditcontent->cr_oids_must, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_MAY_ATTRIBUTES ) == 0)
            return new_values(&element->element.ditcontent->cr_oids_may, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NOT_ATTRIBUTES ) == 0)
            return new_values(&element->element.ditcontent->cr_oids_not, 
                              values, modType);
        p_xterms = &element->element.ditcontent->cr_extensions;
        break;
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:            
        if (strcasecmp(fieldName, LDAP_SCHEMA_RULE_ID ) == 0)
            return new_value(&element->element.ditstructure->sr_rule_id, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_DESCRIPTION ) == 0)
            return new_value(&element->element.ditstructure->sr_desc, 
                             values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAMES ) == 0)
            return new_values(&element->element.ditstructure->sr_names, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_OBSOLETE ) == 0) {
            /* no value */                
            element->element.ditstructure->sr_obsolete = 
            (modType == LDAP_MOD_DELETE) ? 0 : 1;
            return LDAP_SUCCESS;
        }
        else if (strcasecmp(fieldName, LDAP_SCHEMA_NAME_FORM_OBJECT ) == 0)
            return new_value(&element->element.ditstructure->sr_nf_oid, 
                              values, modType);
        else if (strcasecmp(fieldName, LDAP_SCHEMA_SUPERIORS ) == 0)
            return new_values(&element->element.ditstructure->sr_sup_ids, 
                              values, modType);
        p_xterms = &element->element.ditstructure->sr_extensions;
        break;
    default:
        assert(0);
        return LDAP_PARAM_ERROR;
    }         

    /* Non-standard fields: */

    if (*p_xterms != NULL) {
        for (i=0; (*p_xterms)[i] != NULL; i++)
            if ( strcasecmp( (*p_xterms)[i]->lsei_name, fieldName) == 0 )
                return new_values( &((*p_xterms)[i]->lsei_values), 
                                   values, modType );

            /* if we get here we couldn't find the field name */
        if (modType == LDAP_MOD_DELETE)
            return LDAP_SUCCESS;

        /*  Note: i equals the index of the last element (NULL). 
            We only add one new field */
        newSize = i+2;  
    } else
        newSize = 2;    /* create a new list */

    /* Here we create a new field and we either resize or create a list */

    temp = LDAP_REALLOC( *p_xterms, 
               newSize * sizeof( LDAPSchemaExtensionItem *));
    if (temp == NULL) 
        return LDAP_NO_MEMORY;
    else 
        *p_xterms = temp;
    
    p_xterm = LDAP_CALLOC( 1, sizeof(LDAPSchemaExtensionItem )); 
    if (p_xterm == NULL) return LDAP_NO_MEMORY;
    
    p_xterm->lsei_name = LDAP_STRDUP(fieldName);
    if (p_xterm->lsei_name == NULL) {
        LDAP_FREE( p_xterm );
        return LDAP_NO_MEMORY;
    }
    
    p_xterm->lsei_values = str_array_dup(values);
    if (p_xterm->lsei_values == NULL) {
        LDAP_FREE( p_xterm->lsei_name);
        LDAP_FREE( p_xterm );
        return LDAP_NO_MEMORY;
    }
    
    (*p_xterms)[ newSize - 2 ] = p_xterm;
    (*p_xterms)[ newSize - 1 ] = NULL;
    return LDAP_SUCCESS;

}



/*********************************************************************
    remove_element:
    Retrieves a schema element from the list specified by elementType 
    according to the oid passed in. Removes references from both the names
    and oids list.  This removes all references, and deletes the 
    element using free_element.

    Uses a binary search - assumes the lists are sorted. Then re-sorts them.
*/
int remove_element(
                  LDAPSchema* schema, 
                  char* oid, 
                  int elementType )
{
    LDAPSchemaElement **element=NULL;
    struct ldap_schema_named_element *namedEl;
    char **name = NULL;
    int len = 0 , i = 0;
    if (schema->testInit != KEYWORD)
        return LDAP_LOCAL_ERROR;

    if ((oid == NULL) || (schema == NULL)       ||
        (elementType < LDAP_SCHEMA_ATTRIBUTE_TYPE)  ||
        (elementType > LDAP_SCHEMA_DIT_STRUCTURE_RULE) )
        return LDAP_PARAM_ERROR;

    /* oid lookup */
    element = (LDAPSchemaElement**)
              bsearch((void*) oid,                        /* key  */
                      (void*) schema->oids[elementType],  /* list */
                      schema->oidCount[elementType],                   
                      sizeof( LDAPSchemaElement *),
                      compare_key_to_oid  );
    if (element == NULL)
        return LDAP_NO_SUCH_OBJECT;

    switch ((*element)->type) {
    case LDAP_SCHEMA_ATTRIBUTE_TYPE:
        name = str_array_dup( (*element)->element.attribute->at_names );
        break;
    case LDAP_SCHEMA_OBJECT_CLASS:
        name = str_array_dup( (*element)->element.object->oc_names );
        break;
    case LDAP_SCHEMA_SYNTAX:
        name = str_array_dup( (*element)->element.syntax->syn_names );
        break;
    case LDAP_SCHEMA_MATCHING_RULE:
        name = str_array_dup( (*element)->element.matching->mr_names );
        break;
    case LDAP_SCHEMA_MATCHING_RULE_USE:
        name = str_array_dup( (*element)->element.matchinguse->mru_names );
        break;
    case LDAP_SCHEMA_NAME_FORM:
        name = str_array_dup( (*element)->element.nameform->nf_names );
        break;
    case LDAP_SCHEMA_DIT_CONTENT_RULE:
        name = str_array_dup( (*element)->element.ditcontent->cr_names );
        break;
    case LDAP_SCHEMA_DIT_STRUCTURE_RULE:
        name = str_array_dup( (*element)->element.ditstructure->sr_names );
        break;
    default:
        return LDAP_PARAM_ERROR;
    }

    free_element( *element );
    *element = NULL; /* this will get sorted to the bottom */

    len = ldap_count_values(name);
    for ( i=0; i< len; i++) {
        /* name lookup */
        namedEl  = (struct ldap_schema_named_element *)
                   bsearch((void*) name[i],                     /* key  */
                           (void*) schema->names[elementType],  /* list */
                           schema->nameCount[elementType],                   
                           sizeof( struct ldap_schema_named_element ),
                           compare_key_to_name );
        if (namedEl == NULL)
            continue;
        namedEl->p_element = NULL;
        LDAP_FREE(namedEl->name);
        namedEl->name = NULL;
    }

    /* re-sort */
    qsort(  (void*)schema->oids[elementType],       /* list to sort         */
            schema->oidCount[elementType],          /* number of elements   */
            sizeof( LDAPSchemaElement*),            /* size of each item    */
            compare_oid  );                         /* Comparison function  */
    qsort(  (void*)schema->names[elementType],
            schema->nameCount[elementType],
            sizeof( struct ldap_schema_named_element ),
            compare_name );
    schema->nameCount[elementType] -= len;
    schema->oidCount [elementType] -= 1;
    LDAP_VFREE(name);
    return LDAP_SUCCESS;
}



/*********************************************************************
    new_value:
    Used to create a value for single valued fields.  It frees the 
    current value in p_str, sets it to NULL or creates a copy of 
    the new value.  It assumes that values is an array of strings 
    with one string in it.  Used by modify_element.
 */
int new_value( char** p_str, char *values[], int modType)
{
    if (*p_str != NULL)
        LDAP_FREE(*p_str);
    if ((values == NULL) || (values[0] == NULL) || 
        (modType == LDAP_MOD_DELETE))
        *p_str = NULL;
    else {
        *p_str = LDAP_STRDUP( values[0] );
        if (*p_str == NULL)
            return LDAP_NO_MEMORY;
    }
    return LDAP_SUCCESS;
}

/*********************************************************************
    new_values:
    Used to create a value for multi-valued fields.  It creates the 
    new field from p_old and values according to modType.  Used by 
    modify_element.
 */

int new_values( char *(*p_old[]), char *values[], int modType)
{
    char **newValues = NULL;
    int i = 0,j = 0, countNew = 0, 
    countOld    = ldap_count_values(*p_old),
    countValues = ldap_count_values(values);

    if ( modType == LDAP_MOD_REPLACE )
        newValues = str_array_dup( values );
    else {
        /* Remove duplicate values, for both add and delete */       
        countNew = countOld; 
        for (i=0; i < countOld; i++)
            for (j=0; j < countValues; j++)
                /* delete values that match */
                if ( strcasecmp( (*p_old)[i], values[j]) == 0) {
                    LDAP_FREE( (*p_old)[i] );
                    (*p_old)[i] = NULL;
                    countNew --;
                };        
        if ( modType == LDAP_MOD_ADD )
            countNew += countValues;

        /* countNew doesn't include NULL */
        newValues = LDAP_CALLOC( countNew + 1, sizeof( char* ));
        for (i=0, j=0; i< countOld; i++)
            if ((*p_old)[i] != NULL) {   
                /* if the value wasn't deleted, put it in the new list */
                newValues[j] = (*p_old)[i];
                j++;
                (*p_old)[i] = NULL;
            };        

        if ( modType == LDAP_MOD_ADD ) { /* add values */            
            for ( j=0, i = countNew - countValues; i < countNew; i++,j++){
                newValues[i] = LDAP_STRDUP(values[j]);
                if (newValues[i] == NULL) {
                    LDAP_VFREE(*newValues);
                    return LDAP_NO_MEMORY;
                }
            }
        }
        newValues[countNew] = NULL;
    }
    if ( *p_old != NULL )
        LDAP_VFREE(*p_old);
    *p_old = newValues;

    return LDAP_SUCCESS;
}


/**********************************************************************
    add_change

    Adds a change to the list of changes.  These changes will be sent
    to the directory when ldap_schema_save is called.  Used by store_update,
    and ldap_schema_modify.
*/

int add_change(LDAPSchema *schema, LDAPMod *change){
    LDAPMod **temp, *newMod;
    if (schema->changesCount + 1 >= schema->changesAllocated){
        temp = LDAP_REALLOC(schema->changes, 
            (schema->changesAllocated + 10) * sizeof(LDAPMod *));
        if (temp == NULL)
            return LDAP_NO_MEMORY;
        schema->changes = temp;
        schema->changesAllocated += 10;
    }
    /* duplicate all memory for change */
    newMod = LDAP_MALLOC( sizeof(LDAPMod) );
    if (newMod == NULL)
        return LDAP_NO_MEMORY;

    newMod->mod_values  = str_array_dup( change->mod_values );

    if ((change->mod_values != NULL) && 
        (newMod->mod_values == NULL))
        return LDAP_NO_MEMORY;
    
    newMod->mod_op      = change->mod_op;
    newMod->mod_type    = change->mod_type; 
    /* mod_type has a #define string so we don't need to duplicate this */
    
    schema->changes[schema->changesCount] = newMod;
    schema->changesCount ++;
    schema->changes[schema->changesCount] = NULL;
    return LDAP_SUCCESS;
}

/**********************************************************************
    ldap_schema_save

    All changes on schema from ldap_schema_add, modify and delete will
    be saved in the array of LDAPMod structures.  This function sends 
    the LDAPModifications to the directory connected to by the session
    handle and will write the changes to the subschemaSubentry passed in.
    If the subschemaSubentry is NULL then the first subschemaSubentry in
    the Root DSE will be used.

*/
LDAP_F( int )
ldap_schema_save( LDAP *ld, LDAPSchema *schema, const char *subschemaDN )
{
    int rc = 0;
    char* schemaDN = NULL;
    if ((ld == NULL ) || ( schema == NULL ) ||
        ( schema->testInit != KEYWORD ) )
        return LDAP_PARAM_ERROR;
    
    if ((schema->changes == NULL) || (schema->changes[0] == NULL))
        return LDAP_SUCCESS;
    if (subschemaDN == NULL){
        rc = get_schema_dn( ld, &schemaDN );
        if (rc != LDAP_SUCCESS)
            return rc;
        rc = ldap_modify_ext_s( ld, schemaDN, schema->changes, NULL, NULL);
        LDAP_FREE(schemaDN);
    }
    else
        rc = ldap_modify_ext_s( ld, subschemaDN, schema->changes, NULL, NULL);
    
    free_changes(schema);
    return rc;
}


/**********************************************************************
  free_changes:

	Free changes will free any changes stored in the 'changes' list.
	Memory in this list is allocated by ldap_schema_save and consists
	of the list itself, individual mod structures, and the values inside
	of the mod structures.
*/
void free_changes(LDAPSchema *schema){
    int i;
 	if (schema->changes != NULL){
		for (i = 0; schema->changes[i] != NULL; i++){
			if (schema->changes[i]->mod_values != NULL)
				LDAP_VFREE(	schema->changes[i]->mod_values );
			LDAP_FREE( schema->changes[i] );
		}
		LDAP_FREE( schema->changes );
        schema->changes = NULL;
	}   
    schema->changesCount = 0;
    schema->changesAllocated = 0;
}

