/* $Novell: globals.c,v 1.45 2002/10/17 16:40:42 $
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 *****************************************************************************
 * Copyright (C) 1999, 2000 Novell, Inc. All Rights Reserved.
 *
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 *****************************************************************************/


/****************************************************************************
*
* Description:   global variables for the SSL Library
*
****************************************************************************/

#include <stdio.h>


#if defined(N_PLAT_WIN32) || defined(N_PLAT_WIN64)
   #include <windows.h>
   
#elif defined(MODESTO)
   #include "nksapi.h"
#endif

#include "ldap.h"

#ifdef UNIX
#include "openssl/ssl.h"  /* UNIX uses forward slash for path names*/
#include <pthread.h>
#include <sys/errno.h>
#else
#include "openssl\ssl.h" 
#endif 

#include "ssl_io.h"
#include "globals.h"




/********************************************************************
  Win32 platform specific globals                                 
  
   g_KeyFile   - cert filename NILE SSL 
   
   g_CryptCtx      - master ssl crypto ctx 
   g_MaxSockets    - size of g_pSockets array
   g_Locks         - array of mutexes used for OpenSSL locks
   g_pCertCallback - user registered verification callback
   
   g_SASId     - unique number required by sas apis
   
   g_SSLReg    - flag indicated if ldapssl_client_init was called
   g_Sync      - critical section use to sync accessing globals
    
*********************************************************************/
#if defined (N_PLAT_WIN32) || defined(N_PLAT_WIN64)

   SSL_CTX           *g_CryptCtx      = NULL;
   unsigned int       g_MaxSockets    = 0;
   ldapssl_Socket    *g_pSockets      = NULL;
   HANDLE            *g_Locks         = NULL;
   void              *g_pCertCallback = NULL;
   
   int                g_SSLReg = 0;

   CRITICAL_SECTION   g_Sync;


/********************************************************************/
/* Modesto platform globals (see Win32 section for variable desc.   */
/********************************************************************/
#elif defined (MODESTO)

   SSL_CTX           *g_CryptCtx      = NULL;
   unsigned int       g_MaxSockets    = 0;
   ldapssl_Socket    *g_pSockets      = NULL;
   NXMutex_t        **g_Locks         = NULL;
   void              *g_pCertCallback = NULL;

   int                g_SSLReg = 0;
   NXMutex_t         *g_Sync   = NULL;


/********************************************************************/
/* Unix platform globals                                            */
/********************************************************************/
#elif defined(UNIX)

   
   SSL_CTX           *g_CryptCtx     = NULL;
   unsigned int       g_MaxSockets   = 0;
   ldapssl_Socket    *g_pSockets     = NULL;
   void              *g_pCertCallback = NULL;
   void              *g_pCertAlertCB = NULL;
   pthread_mutex_t    g_Locks[CRYPTO_NUM_LOCKS];
   int g_SSLReg = 0;
   
   pthread_mutex_t g_Sync = PTHREAD_MUTEX_INITIALIZER;

   #ifdef HAVE_SOLARIS
     #pragma init (LockInit)
     #pragma fini (LockCleanUp)
   #endif



#endif /* defined(N_PLAT_WIN32) */




/********************************************************************/
/*  UNIX  platform specific functions                               */
/********************************************************************/
#ifdef UNIX

void LockCallback(int mode, int type, const char *file, int line)
{
   if (mode & CRYPTO_LOCK)
      pthread_mutex_lock(&(g_Locks[type]));
   else
      pthread_mutex_unlock(&(g_Locks[type]));
   return;
}

void  LockInit(void)
{
   int i = 0;

   for (i=0; i<CRYPTO_NUM_LOCKS; i++)
     pthread_mutex_init(&(g_Locks[i]),NULL);

   CRYPTO_set_id_callback((unsigned long (*)(void)) pthread_self);
   CRYPTO_set_locking_callback(LockCallback);
   return;
}

void LockCleanUp(void)
{
   int i = 0;

   for (i = 0 ; i < CRYPTO_NUM_LOCKS; i++)
   {
       if (pthread_mutex_trylock(&(g_Locks[i])) == EBUSY)
	 pthread_mutex_unlock(&(g_Locks[i]));
       pthread_mutex_destroy(&(g_Locks[i]));
   }
return ;
}

#endif  /* ifdef UNIX */


/********************************************************************/
/*  WIN32 platform specific functions                               */
/********************************************************************/

#if defined(N_PLAT_WIN32) || defined(N_PLAT_WIN64)

   /* Most of these are defines in globals.h
   */

/********************************************************************
 * Callback used by OpenSSL when aquiring a specific lock - type
 * is the index into the array of mutexes
*******************************************************************/
void LockCallback(int mode, int type, const char *file, int line)
{
   if (mode & CRYPTO_LOCK)
      WaitForSingleObject(g_Locks[type],INFINITE);
   else
      ReleaseMutex(g_Locks[type]);
   return;
}

/********************************************************************
 * Initializes the Mutexes used by OpenSSL and sets the locking
 * callback
*******************************************************************/
int LockInit(void)
{
   int i = 0;

   g_Locks=malloc( CRYPTO_num_locks() * sizeof(HANDLE) );

   if (NULL == g_Locks)
      return(0);

   for (i=0; i<CRYPTO_num_locks(); i++)
      g_Locks[i]=CreateMutex(NULL, FALSE, NULL);

   CRYPTO_set_locking_callback(LockCallback);
   
   return(1);
}

/********************************************************************
 * Cleans up the Mutexes used by OpenSSL
*******************************************************************/
void LockCleanUp(void)
{
   int i = 0;

   CRYPTO_set_locking_callback(NULL);

   for (i=0; i<CRYPTO_num_locks(); i++)
      CloseHandle(g_Locks[i]);

   free(g_Locks);
   g_Locks = NULL;

   return;
}

#endif /* N_PLAT_WIN32 */


/********************************************************************/
/*  Modesto platform specific functions                             */
/********************************************************************/

#if defined(MODESTO)

   /* Most of these are defines in globals.h
   */

/********************************************************************
 * Callback used by OpenSSL when aquiring a specific lock - type
 * is the index into the array of mutexes
*******************************************************************/
void LockCallback(int mode, int type, const char *file, int line)
{
   if (mode & CRYPTO_LOCK)
      NXLock(g_Locks[type]);
   else
      NXUnlock(g_Locks[type]);
   return;
}

/********************************************************************
 * Initializes the Mutexes used by OpenSSL and sets the locking
 * callback
*******************************************************************/
int LockInit(void)
{
   int i = 0;
   NXLockInfo_t info = {"LDAPSSL:OpenSSL Mutex", 0, {0,0}};

   g_Locks = malloc( CRYPTO_num_locks() * sizeof(NXMutex_t*) );

   if (NULL == g_Locks)
      return(0);

   for (i=0; i<CRYPTO_num_locks(); i++)
      g_Locks[i]=NXMutexAlloc(0, 0, &info);

   CRYPTO_set_locking_callback(LockCallback);
   
   return(1);
}

/********************************************************************
 * Cleans up the Mutexes used by OpenSSL
*******************************************************************/
void LockCleanUp(void)
{
   int i = 0;
   
   CRYPTO_set_locking_callback(NULL);

   for (i=0; i<CRYPTO_num_locks(); i++)
      NXMutexFree(g_Locks[i]);

   free(g_Locks);
   g_Locks = NULL;
   
   return;
}

#endif /* MODESTO */

