/* $Novell: ssl_cert.c,v 1.37 2002/11/21 17:45:28 $ 
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 *****************************************************************************
 * Copyright (C) 1999, 2000 Novell, Inc. All Rights Reserved.
 * 
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 
  *****************************************************************************/

/****************************************************************************
*
* Description:   This file contains certificate functions 
*
****************************************************************************/

#include <stdio.h>
#include "ldap_ssl.h"

#include "openssl/ssl.h"
#include "openssl/rand.h"
#include "openssl/err.h"
#include "openssl/bio.h"

#if defined(MODESTO)
#include "string.h"
#include "nksapi.h"
#endif

#include "globals.h"
#include "debug.h"
#include "pkcs12x.h"


/****************************************************************************
* Local Function Prototypes - begin
****************************************************************************/
int add_cert
(
   X509 *pX509
);

BIO *file_to_BIO
(
   char* fileName
);

BIO *pkcs12_BIO_to_clear_BIO
(
   BIO*  in,
   char* password
);

int verify_callback
(
   int ok,
   X509_STORE_CTX *ctx
);
/****************************************************************************
* Local Function Prototypes - end
****************************************************************************/

/****************************************************************************
*
* Description:
*
*  ldapssl_get_cert_attribute - Get attribute out of certificate handle.
*
*  caller can pass in NULL for "value" to query the length of the attribute
*
****************************************************************************/
LDAP_F(int) ldapssl_get_cert_attribute
(
   void          *certHandle,
   int           attrID,
   void          *value,
   int           *length
)
{
   int            rc                 = LDAPSSL_SUCCESS;
   ASN1_TIME      *pAns1Time         = NULL;
   int            tmpLength          = 0;
   X509_NAME      *pX509Name         = NULL;
   char           *pTemp             = NULL;
   X509           *pCert             = NULL;
   X509_STORE_CTX *pStoreCtx         = NULL;
   LDAPSSL_Cert_Validity_Period *pVP = NULL;

#ifdef LDAPSSL_DEBUG
   DBG_FUNC("ldapssl_get_cert_attribute");
   DBG_PARM(" certHandle: 0x%p\n", certHandle);
   DBG_PARM(" attrID: %s\n", DbgCertAttr(attrID) );
   DBG_PARM(" value: 0x%p\n", value);
   DBG_PARM(" length: 0x%p\n", length);
   DBG_PARM(" *length: %d\n", length ? *length : 0 );
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/
   
   if (NULL == certHandle)
   {                                                     
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: invalid certHandle parameter\n");            
      DBG_RETURN(LDAPSSL_ERROR);                         
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);                         
   }                                                     
   
   pStoreCtx = (X509_STORE_CTX *)certHandle;
   pCert = X509_STORE_CTX_get_current_cert(pStoreCtx);

   switch (attrID)
   {
      case LDAPSSL_CERT_ATTR_ISSUER:
         pX509Name = X509_get_issuer_name(pCert);

         pTemp = X509_NAME_oneline(pX509Name, NULL, 0);

         tmpLength = (int)strlen(pTemp);

         if (NULL != value)
         {
            strncpy( (char*)value, pTemp, *length);
            if (*length > tmpLength)
               *length = tmpLength;
         }
         else
            *length = tmpLength;

         OPENSSL_free(pTemp);

         break;

      case LDAPSSL_CERT_ATTR_SUBJECT:
         pX509Name = X509_get_subject_name(pCert);

         pTemp = X509_NAME_oneline(pX509Name, NULL, 0);

         tmpLength = (int)strlen(pTemp);

         if (NULL != value)
         {
            strncpy( (char*)value, pTemp, *length);
            
               /* update Length to the # of chars copied
               */
            if (*length > tmpLength) 
               *length = tmpLength;
         }
         else
            *length = (int)strlen(pTemp);

         OPENSSL_free(pTemp);

         break;
         
      case LDAPSSL_CERT_ATTR_VALIDITY_PERIOD:
         pVP = (LDAPSSL_Cert_Validity_Period*)value;

         *length = sizeof(LDAPSSL_Cert_Validity_Period);

         if (NULL != pVP)
            memset(pVP, 0, sizeof(LDAPSSL_Cert_Validity_Period));
         else
            break;

            /* get the not after time
            */
         pAns1Time = X509_get_notAfter(pCert);
         if (NULL != pAns1Time->data)
         {
            strncpy(pVP->notAfterTime, (char*)pAns1Time->data, 
                    sizeof(pVP->notAfterTime) );      

               /* make sure it is NULL terminated
               */
            pVP->notAfterTime[(sizeof(pVP->notAfterTime))-1] = '\0';

               /* assume utc format - but change if generalized format
               */
            pVP->notAfterType = LDAPSSL_CERT_UTC_TIME;

            if (V_ASN1_GENERALIZEDTIME == pAns1Time->type)
               pVP->notAfterType = LDAPSSL_CERT_GEN_TIME;
         }

            /* get the not before time
            */
         pAns1Time = X509_get_notBefore(pCert);
         if (NULL != pAns1Time->data)
         {
            strncpy(pVP->notBeforeTime, (char*)pAns1Time->data, 
                    sizeof(pVP->notBeforeTime) );      

               /* make sure it is NULL terminated
               */
            pVP->notBeforeTime[(sizeof(pVP->notBeforeTime))-1] = '\0';

               /* assume utc format - but change if generalized format
               */
            pVP->notBeforeType = LDAPSSL_CERT_UTC_TIME;

            if (V_ASN1_GENERALIZEDTIME == pAns1Time->type)
               pVP->notBeforeType = LDAPSSL_CERT_GEN_TIME;
         }

         break;


      case LDAPSSL_CERT_GET_STATUS:
            
         tmpLength = sizeof(int);
   
         if (NULL != value)
         {
            if (*length == tmpLength)
              *((int *)value) = (int)(pStoreCtx->error);
            else
               rc = LDAPSSL_ERROR;
         }
         else
            *length = tmpLength;
   
         break;

      default:

#ifdef LDAPSSL_DEBUG
         DBG_OUTPUT(" ERROR: Unknown attrID: %ld\n", attrID);
#endif /*LDAPSSL_DEBUG*/

         rc = LDAPSSL_ERROR;
   }

#ifdef LDAPSSL_DEBUG
   DBG_RETURN(rc);
#endif /*LDAPSSL_DEBUG*/

   return(rc);
}


/****************************************************************************
*
* Description:
*
*  ldapssl_get_cert - returns the certificate formated as specified by 
*   encoding param
****************************************************************************/
LDAP_F(int) ldapssl_get_cert
(
   void          *certHandle,
   int           type,
   LDAPSSL_Cert  *cert
)
{
   BIO   *pBio = NULL;
   int   rc    = LDAPSSL_SUCCESS;
   int   X509rc = 0;
   X509  *pCert = NULL;

#ifdef LDAPSSL_DEBUG
   DBG_FUNC("ldapssl_get_cert");
   DBG_PARM(" certHandle: 0x%p\n", certHandle);
   DBG_PARM(" type: %s\n", DbgCertType(type));
   DBG_PARM(" LDAPSSL_Cert: 0x%p\n", cert);
   DBG_PARM(" LDAPSSL_Cert->length: %0d\n", cert ? cert->length : 0 );
   DBG_PARM(" LDAPSSL_Cert->data: 0x%p\n", cert ? cert->data : NULL );
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/
   
   if(NULL == certHandle || NULL == cert)                       
   {                                                     
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: invalid parameter\n");            
      DBG_RETURN(LDAPSSL_ERROR);
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);
   }
   
   pCert = X509_STORE_CTX_get_current_cert((X509_STORE_CTX *)certHandle);
   
   // Create BIO
   pBio = BIO_new(BIO_s_mem());
   if(NULL == pBio)
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: Cannot allocate BIO memory.\n");
      DBG_RETURN(LDAPSSL_ERROR);
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);
   }

   // Convert certificate to BIO
   if (LDAPSSL_CERT_BUFFTYPE_B64 == type) // PEM or B64
   {
      X509rc = PEM_write_bio_X509_AUX(pBio, pCert);
      if(0 == X509rc)
      {
#ifdef LDAPSSL_DEBUG
         DBG_STRING(" ERROR: PEM_write_bio_X509_AUX returned error.\n");
#endif /*LDAPSSL_DEBUG*/
         rc = LDAPSSL_ERROR;
      }
   }
   else if (LDAPSSL_CERT_BUFFTYPE_DER == type) // ASN1 or DER
   {
      X509rc = i2d_X509_bio(pBio, pCert);
      if(0 == X509rc)
      {
#ifdef LDAPSSL_DEBUG
         DBG_STRING(" ERROR: i2d_X509_bio returned error.\n");
#endif /*LDAPSSL_DEBUG*/
         rc = LDAPSSL_ERROR;
      }
   }
   else
   {
#ifdef LDAPSSL_DEBUG
      DBG_OUTPUT(" ERROR: Unknown type: %ld\n", type);
#endif /*LDAPSSL_DEBUG*/
      rc = LDAPSSL_ERROR;
   }

      /* Applications can query the size of the certificate by setting
       * the LDAPSSL_Cert->data pointer to NULL.  
      */ 
   if (NULL == cert->data)
   {
      cert->length = pBio->num_write;
   }
   else
   {
      cert->length = BIO_read(pBio, cert->data, cert->length);
#ifdef LDAPSSL_DEBUG
      DBG_OUTPUT(" BIO_read returns length of: %ld\n", cert->length);
#endif /*LDAPSSL_DEBUG*/

      if(0 == cert->length)
      {
#ifdef LDAPSSL_DEBUG
         DBG_STRING(" ERROR: BIO_read returned 0.\n");
#endif /*LDAPSSL_DEBUG*/
         rc = LDAPSSL_ERROR;
      }
      else if(-1 == cert->length)
      {
#ifdef LDAPSSL_DEBUG
         DBG_STRING(" ERROR: BIO_read returned -1.\n");
#endif /*LDAPSSL_DEBUG*/
         rc = LDAPSSL_ERROR;
      }
      else if(-2 == cert->length)
      {
#ifdef LDAPSSL_DEBUG
         DBG_STRING(" ERROR: BIO_read returned -2.\n");
#endif /*LDAPSSL_DEBUG*/
         rc = LDAPSSL_ERROR;
      }
   }

   BIO_free(pBio);
   
#ifdef LDAPSSL_DEBUG
   DBG_RETURN(rc);
#endif /*LDAPSSL_DEBUG*/
   return(rc);
}

/****************************************************************************
*
* Description:
*
*  ldapssl_set_verify_callback - set the users certificate callback.
*
*  CertVerifyFunc is a pointer to a function that returns an int and takes
*  one void* param
*
****************************************************************************/
LDAP_F(int) ldapssl_set_verify_callback
(
   int (*certVerifyFunc)(void *)
)
{   
   SSL_CTX *ctx = NULL;
   
#ifdef LDAPSSL_DEBUG
   DBG_FUNC("ldapssl_set_cert_callback");
   DBG_PARM(" certVerifyFunc: 0x%p\n", certVerifyFunc);
#endif /*LDAPSSL_DEBUG*/

   EnterSync();

   if (0 == GET_SSL_REGISTERED)
   {
      ExitSync();
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: SSL library not initialized\n");
      DBG_RETURN(LDAPSSL_ERROR);			  
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);			  
   }
   
   ctx = GetMasterCryptCtx();
   if (NULL == ctx)
   {
      ExitSync();
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: NULL ctx.\n");            
      DBG_RETURN(LDAPSSL_ERROR);			  
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);			  
   }

      /* Set the master crypt context's verify callback function - 
       * all newly created connections will be enabled for interactive
       * verification.  Existing connections will not be effected.
       *
       * NOTE: the callback function is ldapssl's function and it
       * checks the status and determines if the client's verify
       * function should be called.  This is because OpenSSL calls 
       * the verify function for everything (success or failure) but
       * we only want to call the client's function if the certificate
       * is not trusted
      */
   if (NULL != certVerifyFunc)
   {
      SSL_CTX_set_verify( ctx, SSL_VERIFY_PEER,
                          verify_callback);

         /* Set the clients verify function callback which will be called 
          * be verify_callback if necessary
         */
      SetCertCallback((void *)certVerifyFunc);
   }
   else
   {
         /* NULL is passed in to clear the verify callback.
         */
      SSL_CTX_set_verify( ctx, SSL_VERIFY_PEER, NULL ); 
      SetCertCallback(NULL);
   }
   
   ExitSync();

#ifdef LDAPSSL_DEBUG
   DBG_RETURN(LDAPSSL_SUCCESS);
#endif /*LDAPSSL_DEBUG*/
   return(LDAPSSL_SUCCESS);
}

/****************************************************************************
*
* Description:
*
*  ldapssl_add_trusted_cert - add a trusted root certificate
*  (file or buffer) to the list of trusted roots in the current master
*  crypt context.
*
****************************************************************************/
LDAP_F(int) ldapssl_add_trusted_cert
(
   void *cert,
   int   type
)
{
   int     rc       = LDAPSSL_SUCCESS;
   int     numCerts = 0;
   X509    *pX509   = NULL;
   BIO     *pBio    = NULL;

#ifdef LDAPSSL_DEBUG
   DBG_FUNC("ldapssl_add_trusted_cert");
   DBG_PARM(" cert: 0x%p\n", cert);
   DBG_PARM(" type: %s\n", DbgCertType(type));
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/

   if (NULL == cert)
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: NULL cert.\n");            
      DBG_RETURN(LDAPSSL_ERROR);
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);
   }

      /* synchronize access because we will be adding cert to the master
       * crypto context
      */
   EnterSync();

   if (0 == GET_SSL_REGISTERED)
   {
      ExitSync();
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: SSL library not initialized\n");
      DBG_RETURN(LDAPSSL_ERROR);
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);
   }
   
      /* Create new BIO based on type - File BIO for files and memory 
       * BIO for buffers
      */
   if (LDAPSSL_CERT_FILETYPE_B64 == type ||
       LDAPSSL_CERT_FILETYPE_DER == type )
   {
      pBio = file_to_BIO((char*)cert);
   }
   else if (LDAPSSL_CERT_BUFFTYPE_B64 == type ||
            LDAPSSL_CERT_BUFFTYPE_DER == type )
   {
      pBio = BIO_new_mem_buf( ((LDAPSSL_Cert*)cert)->data, 
                              ((LDAPSSL_Cert*)cert)->length );
   }
   else
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING("  ERROR: Invalid cert type\n");
#endif /*LDAPSSL_DEBUG*/
   }

      /* can not continue if bio creation failed
      */
   if (NULL == pBio)
   {
      ExitSync();
#ifdef LDAPSSL_DEBUG
      DBG_STRING("  ERROR: BIO creation failed\n");
      DBG_RETURN(LDAPSSL_ERROR);
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);
   }
      

      /* Load the certs - NOTE: B64 type can contain multiple certs so 
       * a loop is used reading one cert at a time until we don't get
       * any more.  Count the number of certs added and if don't 
       * have at least 1 valid cert return an error
      */
   switch (type)
   {
      case LDAPSSL_CERT_FILETYPE_B64:
      case LDAPSSL_CERT_BUFFTYPE_B64:
         for (;;)
         {
            ERR_clear_error();
	    pX509 = PEM_read_bio_X509_AUX(pBio, NULL, NULL, NULL);
            if (NULL == pX509)
            {
               if ((ERR_GET_REASON(ERR_peek_error()) ==
                  PEM_R_NO_START_LINE) && (0 < numCerts))
               {
                  ERR_clear_error();
                  rc = LDAPSSL_SUCCESS;
               }
               else
                  rc = LDAPSSL_ERROR;
               break;
            }
            rc = add_cert(pX509);

            X509_free(pX509);
            pX509 = NULL;
            
	    if(LDAPSSL_ERROR == rc) break;
            
            numCerts++;
         }
         break;

      case LDAPSSL_CERT_FILETYPE_DER:
      case LDAPSSL_CERT_BUFFTYPE_DER:
         ERR_clear_error();
         X509_free(pX509);

         pX509 = d2i_X509_bio(pBio, NULL);
         if (NULL == pX509)
         {
#ifdef LDAPSSL_DEBUG
            DBG_STRING("  ERROR: d2i_X509_bio returned NULL\n");
#endif /*LDAPSSL_DEBUG*/
            rc = LDAPSSL_ERROR;
            break;
         }         
         rc = add_cert(pX509);
         X509_free(pX509);
         pX509 = NULL;
         
         break;

      default:
#ifdef LDAPSSL_DEBUG
         DBG_STRING("  ERROR: unknown cert type\n");
#endif /*LDAPSSL_DEBUG*/
         rc = LDAPSSL_ERROR;
         break;
   }

   ExitSync();

      /* these functions check for NULL so we won't bother to check
      */
   X509_free(pX509);
   BIO_free(pBio);

#ifdef LDAPSSL_DEBUG
   DBG_RETURN(rc);
#endif /*LDAPSSL_DEBUG*/
   return(rc);
}
   

/****************************************************************************
*
* Description:
*
*  ldapssl_set_client_private_key - Insert the client private key into the
*  SSL context. This causes SSL to use the proper private key for encrypting. 
****************************************************************************/
LDAP_F(int) ldapssl_set_client_private_key
(
   void *key,
   int   keyType,
   void *keyPassword
)
{
   int      rc    = LDAP_SUCCESS;
   int      SSLrc = 0;
   SSL_CTX *ctx   = NULL;

#ifdef LDAPSSL_DEBUG
   DBG_FUNC("ldapssl_set_client_private_key");
   DBG_PARM(" key: 0x%p\n", key);
   DBG_PARM(" keyType: %s\n", DbgEncodingType(keyType));
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/

   /* Synchronize access because we will be adding user cert and private key
    * to the master crypto context.
    */
   EnterSync();

   if (NULL == key)
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: NULL key.\n");
#endif /*LDAPSSL_DEBUG*/
      rc = LDAP_PARAM_ERROR;
      goto CleanUp;
   }
   
   if (0 == GET_SSL_REGISTERED)
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: SSL library not initialized\n");
#endif /*LDAPSSL_DEBUG*/
      rc = LDAP_OPERATIONS_ERROR;
      goto CleanUp;
   }
   
   ctx = GetMasterCryptCtx();
   if (NULL == ctx)
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: NULL ctx.\n");
#endif /*LDAPSSL_DEBUG*/
      rc = LDAP_OPERATIONS_ERROR;
      goto CleanUp;
   }

   /* Set the clients password in the SSL Context
    */
   SSL_CTX_set_default_passwd_cb_userdata(ctx, keyPassword);
   

   switch (keyType)
   {
      /* Use B64 (PEM) file as the private key in the SSL context.
       */
      case LDAPSSL_FILETYPE_B64:
      {
         SSLrc = SSL_CTX_use_PrivateKey_file(ctx, (char*)key,
                                             SSL_FILETYPE_PEM);
         break;
      }   

      /* Use the B64 (PEM) buffer or PKCS12 buffer/file as the private key
       * in the SSL context.
       */
      case LDAPSSL_BUFFTYPE_B64:
      case LDAPSSL_FILETYPE_P12:
      case LDAPSSL_BUFFTYPE_P12:
      {
         BIO *in        = NULL;
         BIO *out       = NULL;
         EVP_PKEY *pkey = NULL;

         if (LDAPSSL_FILETYPE_P12 == keyType)
         {
            /* Read file into BIO.
             */
             in = file_to_BIO((char *)key);
         }
         else
         {
            /* Read the buffer into a BIO.
             */
            in = BIO_new_mem_buf(
                 ((LDAPSSL_Private_Key*)key)->data, 
                 ((LDAPSSL_Private_Key*)key)->length );
         }

         if(NULL != in)
         {
            if (LDAPSSL_FILETYPE_P12 == keyType ||
                LDAPSSL_BUFFTYPE_P12 == keyType)
            {
               /* If BIO is in PKCS12 format convert to clear text.
                */
               out = pkcs12_BIO_to_clear_BIO(in, keyPassword);
            }
            else
            {
              /* Already in clear text.
               */
               out = in;
            }
            if (NULL != out)
            {
               /* Get the client private key
                */
               pkey = PEM_read_bio_PrivateKey(out,NULL,
                       ctx->default_passwd_callback,
                       ctx->default_passwd_callback_userdata);
               if(NULL != pkey)
               {
                  /* Insert the client private key into the SSL context
                   */
                  SSLrc = SSL_CTX_use_PrivateKey(ctx,pkey);
               }         
            }
         }
         if (in == out) out = NULL; /* Clean up "out" properly.*/
         if (NULL != in) BIO_free(in);
         if (NULL != out) BIO_free(out);
         if (NULL != pkey) EVP_PKEY_free(pkey);         
         break;
      }
      default:
      {
#ifdef LDAPSSL_DEBUG
         DBG_STRING("  ERROR: Unknown key type.\n");
#endif /*LDAPSSL_DEBUG*/
         SSLrc = 0;
         break;
      }
   }

   if (0 != SSLrc) /* 0 is error in OpenSSL */ 
   {
      rc = LDAP_SUCCESS;
   }
   else
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING("  ERROR: SSL_CTX_use_PrivateKey failed\n");
#endif /*LDAPSSL_DEBUG*/
      rc = LDAP_OPERATIONS_ERROR;
      DBG_OPENSSL_ERROR(rc);
   }

CleanUp:   
   if(NULL != ctx)
      SSL_CTX_set_default_passwd_cb_userdata(ctx, NULL);
   ExitSync();
#ifdef LDAPSSL_DEBUG
   DBG_RETURN(rc);
#endif /*LDAPSSL_DEBUG*/
   return(rc);
}

/****************************************************************************
*
* Description:
*
*  ldapssl_set_client_cert - Set the client certificate for mutual
*  authentication.
****************************************************************************/
LDAP_F(int) ldapssl_set_client_cert
(
   void *cert,
   int   type,
   void *password
)
{
   int      rc    = LDAP_SUCCESS;
   int      SSLrc = 0;
   SSL_CTX *ctx   = NULL;

#ifdef LDAPSSL_DEBUG
   DBG_FUNC("ldapssl_set_client_cert");
   DBG_PARM(" cert: 0x%p\n", cert);
   DBG_PARM(" type: %s\n", DbgEncodingType(type));
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/


   /* Synchronize access because we will be adding user cert to the master
    * crypto context.
    */
   EnterSync();

   if (NULL == cert)
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: NULL cert.\n");            
#endif /*LDAPSSL_DEBUG*/ 
      rc = LDAP_PARAM_ERROR;
      goto CleanUp;
   }

   if (0 == GET_SSL_REGISTERED)
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: SSL library not initialized\n");
#endif /*LDAPSSL_DEBUG*/
      rc = LDAP_OPERATIONS_ERROR;
      goto CleanUp;
   }
   
   ctx = GetMasterCryptCtx();
   if (NULL == ctx)
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: NULL ctx.\n");            
#endif /*LDAPSSL_DEBUG*/
      rc = LDAP_OPERATIONS_ERROR;
      goto CleanUp;
   }
   
   /* Set the clients password in the SSL Context
    */
   SSL_CTX_set_default_passwd_cb_userdata(ctx, password);
      
   switch (type)
   {
   /* Use the DER (ANS1) file as the client certificate in the SSL context.
    */
   case  LDAPSSL_FILETYPE_DER:
   {
      SSLrc = SSL_CTX_use_certificate_file(ctx,
                                           (char*)cert,
                                           SSL_FILETYPE_ASN1);
      break;
   }
  /* Use the DER (ANS1) buffer as the client certificate in the SSL context.
   */
   case LDAPSSL_BUFFTYPE_DER:
   {
      SSLrc = SSL_CTX_use_certificate_ASN1(ctx,
            (int)(((LDAPSSL_Cert*)cert)->length),
            (unsigned char *)(((LDAPSSL_Cert*)cert)->data));
      break;
   }
  /* Use the B64 (PEM)  file as the client certificate in the SSL context.
   */
   case LDAPSSL_FILETYPE_B64:
   {
      SSLrc = SSL_CTX_use_certificate_file(ctx, (char*)cert, SSL_FILETYPE_PEM);
      break;
   }   

  /* Use the B64 (PEM) buffer or PKCS12 buffer/file as the client certificate
   * in the SSL context.
   */
   case LDAPSSL_FILETYPE_P12:
   case LDAPSSL_BUFFTYPE_P12:
   case LDAPSSL_BUFFTYPE_B64:
   {
      BIO *in  = NULL;
      BIO *out = NULL;
      X509 *x  = NULL;

      if (LDAPSSL_FILETYPE_P12 == type)
      {
         /* If file read into BIO.
          */
         in = file_to_BIO((char *)cert);
      }
      else
      {
         /* Read the buffer into a BIO.
          */
         in = BIO_new_mem_buf(
            ((LDAPSSL_Cert*)cert)->data, 
            ((LDAPSSL_Cert*)cert)->length );
      }

      if(NULL != in)
      {
         if (LDAPSSL_FILETYPE_P12 == type ||
             LDAPSSL_BUFFTYPE_P12 == type)
         {
            /* If BIO is in PKCS12 format convert to clear text.
             */
            out = pkcs12_BIO_to_clear_BIO(in, password);
         }
         else
         {
            /* Already in clear text.
             */
              out = in;
         }
         if (NULL != out)
         {
            /* Get the client certificate
             */
            x = PEM_read_bio_X509(out, NULL, ctx->default_passwd_callback,
                                  ctx->default_passwd_callback_userdata);
            if(NULL != x)
            {
               /* Add the client certificate to the SSL Context
                */
               SSLrc = SSL_CTX_use_certificate(ctx,x);
            }         
         }
      }
      if (in == out) out = NULL; /* Clean up "out" properly.*/ 
      if (NULL != in) BIO_free(in);
      if (NULL != out) BIO_free(out);
      if (NULL != x) X509_free(x);
      break;
   }
   default:
      {
#ifdef LDAPSSL_DEBUG
         DBG_STRING("  ERROR: Unknown key type.\n");
#endif /*LDAPSSL_DEBUG*/ 
         SSLrc = 0;
         break;
      }
   }
   
   if (0 != SSLrc) /* 0 is error in OpenSSL */  
   {
      rc = LDAP_SUCCESS;
   }
   else
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING("  ERROR: SSL_CTX_use_certificate failed\n");
#endif /*LDAPSSL_DEBUG*/ 
      rc = LDAP_OPERATIONS_ERROR;
      DBG_OPENSSL_ERROR(rc);
      goto CleanUp;
   }

CleanUp:   
   if(NULL != ctx)
      SSL_CTX_set_default_passwd_cb_userdata(ctx, NULL);
   ExitSync();
#ifdef LDAPSSL_DEBUG
   DBG_RETURN(rc);
#endif /*LDAPSSL_DEBUG*/ 
   return(rc);
}

/****************************************************************************
*
* Description:
*
*  ldapssl_set_verify_mode - sets the verify options for the ssl handshake
*     LDAPSSL_VERIFY_SERVER = verify server certificate
*     LDAPSSL_VERIFY_NONE   = trust everyone, no certificate verification
*
*     NOTE: when using LDAPSSL_VERIFY_NONE, it is not necessary to pass
*     in a certificate file to the ldapssl_client_init function.
*
****************************************************************************/
LDAP_F(int) ldapssl_set_verify_mode
(
   int   mode
)
{
   SSL_CTX *ctx = NULL;

#ifdef LDAPSSL_DEBUG
   DBG_FUNC("ldapssl_set_verify_mode");
   DBG_PARM(" mode: %s\n", DbgVerifyMode(mode));
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/ 

      /* check to make sure a valid mode is specified
      */
   if ( LDAPSSL_VERIFY_NONE != mode && LDAPSSL_VERIFY_SERVER != mode )
#ifdef LDAPSSL_DEBUG
      DBG_RETURN(LDAPSSL_ERROR);
#endif /*LDAPSSL_DEBUG*/ 
      return(LDAPSSL_ERROR);			  


   EnterSync();
   
   if (0 == GET_SSL_REGISTERED)
   {
      ExitSync();
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: SSL library not initialized\n");
      DBG_RETURN(LDAPSSL_ERROR);			  
#endif /*LDAPSSL_DEBUG*/ 
      return(LDAPSSL_ERROR);			  
   }
   
   ctx = GetMasterCryptCtx();
   if (NULL == ctx)
   {
      ExitSync();
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: NULL ctx.\n");            
      DBG_RETURN(LDAPSSL_ERROR);			  
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);			  
   }

      /* NOTE: the ldapssl verify modes match the openSSL modes - if this
       * ever changes we won't be able to directly pass in mode
      */
   SSL_CTX_set_verify( ctx, mode, NULL);
   
   ExitSync();

#ifdef LDAPSSL_DEBUG
   DBG_RETURN(LDAPSSL_SUCCESS);
#endif /*LDAPSSL_DEBUG*/ 
   return(LDAPSSL_SUCCESS);
}

/****************************************************************************
*
* Description:
*
*  ldapssl_get_verify_mode - gets the verify options for the ssl handshake
*     LDAPSSL_VERIFY_SERVER = verify server certificate
*     LDAPSSL_VERIFY_NONE   = trust everyone, no certificate verification
*
****************************************************************************/
LDAP_F(int) ldapssl_get_verify_mode
(
   int   *mode
)
{
   SSL_CTX *ctx = NULL;
   
#ifdef LDAPSSL_DEBUG
   DBG_FUNC("ldapssl_get_verify_mode");
   DBG_PARM(" mode: 0x%p\n", mode);
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/ 

   EnterSync();
   
   if (0 == GET_SSL_REGISTERED)
   {
      ExitSync();
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: SSL library not initialized\n");
      DBG_RETURN(LDAPSSL_ERROR);
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);			  
   }
   ctx = GetMasterCryptCtx();
   if (NULL == ctx)
   {
      ExitSync();
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: NULL ctx.\n");            
      DBG_RETURN(LDAPSSL_ERROR);			  
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);			  
   }

      /* NOTE: the ldapssl verify modes match the openSSL modes - if this
       * ever changes we won't be able to directly return mode
      */
   *mode = SSL_CTX_get_verify_mode( ctx );
   
   ExitSync();

#ifdef LDAPSSL_DEBUG
   DBG_OUTPUT("mode: %s\n", DbgVerifyMode(*mode));
   DBG_RETURN(LDAPSSL_SUCCESS);
#endif /*LDAPSSL_DEBUG*/ 
   return(LDAPSSL_SUCCESS);
}

/****************************************************************************
* Local Functions - begin
****************************************************************************/

/****************************************************************************
*
* Description:
*
*  add_cert - Add a certificate to the cert store  
*
****************************************************************************/
int add_cert
(
   X509 *pX509
)
{
   SSL_CTX *ctx = NULL;

#ifdef LDAPSSL_DEBUG
   DBG_FUNC("add_cert");
   DBG_PARM(" pX509: 0x%p\n", pX509);
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/

   ctx = GetMasterCryptCtx();
   if (NULL == ctx)
   {
#ifdef LDAPSSL_DEBUG
      DBG_STRING(" ERROR: NULL ctx.\n");            
      DBG_RETURN(LDAPSSL_ERROR);			  
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_ERROR);			  
   }
   
   ERR_clear_error();
   if (0 != X509_STORE_add_cert( ctx->cert_store, pX509)) /* 0 is error in OpenSSL */ 
   {
#ifdef LDAPSSL_DEBUG
      DBG_RETURN(LDAPSSL_SUCCESS);			  
#endif /*LDAPSSL_DEBUG*/
      return(LDAPSSL_SUCCESS);			  
   }
   else
   {
         /* don't treat cert already exists as error
         */
      if ( X509_R_CERT_ALREADY_IN_HASH_TABLE == 
           ERR_GET_REASON(ERR_peek_error()) )
      {
#ifdef LDAPSSL_DEBUG
         DBG_STRING("  WARNING: cert already exists\n");
         ERR_clear_error();
         DBG_RETURN(LDAPSSL_SUCCESS);			  
#endif /*LDAPSSL_DEBUG*/
         return(LDAPSSL_SUCCESS);			  
      }
      else
      {
#ifdef LDAPSSL_DEBUG
         DBG_STRING("  ERROR: X509_STORE_add_cert failed\n");
         DBG_RETURN(LDAPSSL_ERROR);			  
#endif /*LDAPSSL_DEBUG*/
         return(LDAPSSL_ERROR);			  
      }
   }
}

/****************************************************************************
*
* Description:
*
*  file_to_BIO - Read file into BIO.
*  Note: return value needs to be freed by BIO_free()
****************************************************************************/
BIO *file_to_BIO
(
   char* fileName
)
{
    BIO  *out = NULL;
    int   rc  = 0;

#ifdef LDAPSSL_DEBUG
   DBG_FUNC("file_to_BIO");
   DBG_PARM(" fileName: %s\n", fileName);
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/
    /* Open a BIO
     */
    out = BIO_new(BIO_s_file());
    if(NULL != out)
    {
       /* Read the pkcs12 file into the BIO
        */
       if(0 >= BIO_read_filename(out, fileName))
       {
           BIO_free(out);
           out = NULL;
       }
    }    
#ifdef LDAPSSL_DEBUG
    DBG_RETURN(out);
#endif /*LDAPSSL_DEBUG*/
    return(out);
}

/****************************************************************************
*
* Description:
*
*  pkcs12_BIO_to_clear_BIO - Read the pkcs12 BIO into a clear text buffer.
*  Note: return value needs to be freed by BIO_free()
****************************************************************************/
BIO *pkcs12_BIO_to_clear_BIO
(
   BIO*  in,
   char* password
)
{
    BIO  *out = NULL;
    int   rc  = 0;

#ifdef LDAPSSL_DEBUG
    DBG_FUNC("pkcs12_BIO_to_clear_BIO");
    DBG_PARM(" in: 0x%p\n", in);
    DBG_PARM(" password: %s\n", password);
    DBG_LINE();
#endif /*LDAPSSL_DEBUG*/

    /* Create the output BIO
     */
    out = BIO_new(BIO_s_mem());
    if(NULL != out)
    {
        /* Dump the pkcs12 BIO to output BIO
         */
        rc = dump_certs_keys_buf(out, in, password);
        if (0 == rc)
        {
            BIO_free(out);
            out = NULL;
        }
    }
#ifdef LDAPSSL_DEBUG
    DBG_RETURN(out);
#endif /*LDAPSSL_DEBUG*/ 
    return(out);
}

/****************************************************************************
*
* Description:
*
*  verify_callback - The internal verify callback routine registered
*  with OpenSSL when an application calls ldapssl_set_verify_callback.  This
*  routine checks the status of the certificate and only calls the 
*  application callback if there is an error.  
*
****************************************************************************/
int verify_callback
(
   int ok,
   X509_STORE_CTX *ctx
)
{
   int   status    = 0;
   int   openSslRc = 1;

   int (*AppVerifyFunc)(void *) = NULL;
   
#ifdef LDAPSSL_DEBUG
   DBG_FUNC("verify_callback");
   DBG_PARM(" ok: %d\n", ok);
   /* DBG_CERT(ctx) - make a function to dump cert info */
   DBG_LINE();
#endif /*LDAPSSL_DEBUG*/

   EnterSync();
   AppVerifyFunc = (int*)(void *)GetCertCallback(); 
   ExitSync();

      /* if there is no user callback or the ctx is NULL, just
       * return
      */
   if(NULL == AppVerifyFunc || NULL == ctx)
   {
#ifdef LDAPSSL_DEBUG
      DBG_RETURN(ok);
#endif /*LDAPSSL_DEBUG*/
      return(ok);
   }
      /* 0 from OpenSSL means that there is something wrong with the 
       * certificate
       */
   if (0 == ok) 
   {
#ifdef LDAPSSL_DEBUG
      DBG_OUTPUT("Certificate Error Number: %d\n", ctx->error);
      DBG_OUTPUT("Certificate Error Depth: %d\n", ctx->depth);
      DBG_OUTPUT("Certificate Error Description: %s\n", X509_verify_cert_error_string(ctx->error));
#endif /*LDAPSSL_DEBUG*/
         /* Call the registered verify function
         */
      status = AppVerifyFunc(ctx); 
#ifdef LDAPSSL_DEBUG
      DBG_OUTPUT("Certificate Alert Callback return status: %d\n", status);
#endif /*LDAPSSL_DEBUG*/
   
      if(LDAPSSL_CERT_ACCEPT == status)
      {
         ok = 1;  /* openssl uses 1 as success */
      }
   }
#ifdef LDAPSSL_DEBUG
   DBG_RETURN(ok);
#endif /*LDAPSSL_DEBUG*/
   return(ok);
}

/****************************************************************************
* Local Functions - end
****************************************************************************/



